!{\src2tex{textfont=tt}}
!!****f* ABINIT/boundy
!! NAME boundy
!!
!!
!! FUNCTION
!! Finds the boundary of the basis sphere of G vectors (at a given
!! k point) for use in improved zero padding of ffts in 3 dimensions.
!! Provides data to be used by subroutine fourwf
!! Usually, provides integer values g3min,g3max, and then for g3 in the sequence
!! g3=0,1,2,...,g3max,g3min,g3min+1,...,-1, provides g2min, g2max.
!! For mod(fftalg,10)==2), provides 
!! integer values g2min,g2max, and then for g2 in the
!! sequence g2=0,1,2,...,g2max,g2min,g2min+1,...,-1, provides g1min, g1max.
!! These values are stored in the above order in array gbound.
!! Debug mode, if fftalg is between 000 and 099
!!
!! COPYRIGHT
!! Copyright (C) 1998-2001  ABINIT group (DCA, XG, GMR)
!! This file is distributed under the terms of the 
!! GNU General Public License, see ~ABINIT/Infos/copyright
!! or http://www.gnu.org/copyleft/gpl.txt .
!! For the initials of contributors, see ~ABINIT/Infos/contributors .
!!
!! INPUTS
!!  fftalg=defines the fft algorithm
!!  ikg=shift to be given to the location of the output data in the array ikg
!!  istwf_k=option parameter that describes the storage of wfs
!!  kg(3,npw)=integer coordinates of G vectors in basis sphere
!!  mgfft=maximum size of 1D FFTs
!!  mkmem =maximum number of k points which can fit in core memory
!!  mpw=maximum number of planewaves as dimensioned in calling routine
!!  npw=number of G vectors in basis at this k point

!!
!! OUTPUT
!!  gbound(2*mgfft+4)=defined above
!!  ngb=final number of elements in gbound
!!
!! SIDE EFFECTS
!!
!!
!! NOTES
!!
!!
!! PARENTS
!!      kpgio,wfconv
!!
!! SOURCE


 subroutine boundy(fftalg,gbound,ikg,istwf_k,kg,mgfft,mkmem,mpw,ngb,npw) 

 use basis_defs
 implicit none

!Arguments ------------------------------------
 integer :: fftalg,ikg,istwf_k,mgfft,mkmem,mpw,ngb,npw
 integer :: gbound(2*mgfft+4),kg(3,mpw*mkmem)

!Local variables-------------------------------
 integer :: dim_a,dim_b,fftalgc,&
& g_a,gmax_a,gmin_a,gmax_b,gmax_b1,gmax_b2,gmin_b,gmin_b1,gmin_b2,&
& ig,igb,igbinv,ikgn,ikg1,iloop,ipw,testm,testp
 integer :: gmax(2),gmin(2)
 character*500 :: message
 
! *************************************************************************
!
!DEBUG
!write(6,*)' boundy : enter, ikg= ',ikg
!write(6, '(a)' )' boundy : list of plane waves coordinates for k point '
!write(6, '(a)' )'       ipw       kg(1:3,ipw+ikg) '
!do ipw=1,npw
! write(6, '(i10,a,3i6)' )ipw,'  ',kg(1:3,ipw+ikg)
!enddo
!ENDDEBUG

 fftalgc=mod(fftalg,10)
 if(fftalgc/=2)then
  dim_a=3
  dim_b=2
 else
  dim_a=2
  dim_b=1
 endif

 ikg1=ikg+1
 ikgn=ikg+npw

!First find gmin_a and gmax_a
 gmin_a=minval(kg(dim_a,ikg1:ikgn))
 gmax_a=maxval(kg(dim_a,ikg1:ikgn))
 gbound(1)=gmin_a
 gbound(2)=gmax_a

!Must complete the sphere for fftalgc==1 and special storage modes.
!Explanation : sg_fftpad is not able to take into account
!the time-reversal symmetry, so that the boundaries will not be delimited
!by the kg set, but by their symmetric also.
 if(istwf_k>=2 .and. fftalgc==1)then
  if( istwf_k==2 .or. istwf_k==3 .or. istwf_k==6 .or. istwf_k==7 )then
   gbound(2)=max(gmax_a,-gmin_a)
!  XG 010419 corrected bug
!  gbound(1)=-gmax_a
   gbound(1)=-gbound(2)
  else if( istwf_k==4 .or. istwf_k==5 .or. istwf_k==8 .or. istwf_k==9 )then
   gbound(2)=max(gmax_a,-gmin_a-1)
!  XG 010419 corrected bug
!  gbound(1)=-gmax_a-1
   gbound(1)=-gbound(2)-1
  endif
 endif
  
!DEBUG
!write(6,*)' boundy : gmin_a and gmax_a '
!write(6,*)gmin_a,gmax_a
!if(.true.)stop
!ENDDEBUG

 igb=3 
!
!Consider first every g_a in range 0 ... gbound(2), then gbound(1) ... -1
 gmin(1)=0         ; gmax(1)=gbound(2)
 gmin(2)=gbound(1) ; gmax(2)=-1

 do iloop=1,2

 if( gmin(iloop) <= gmax(iloop) )then

  do g_a=gmin(iloop),gmax(iloop)

   if(istwf_k==1 .or. fftalgc/=1)then

!   Usual case
    gmin_b=minval(kg(dim_b,ikg1:ikgn), MASK= (kg(dim_a,ikg1:ikgn)==g_a) )
    gmax_b=maxval(kg(dim_b,ikg1:ikgn), MASK= (kg(dim_a,ikg1:ikgn)==g_a) )

   else if(istwf_k>=2 .and. fftalgc==1)then
!   Here, must take into account time-reversal symmetry explicitely

!   Determine the boundaries for the plane g_a
    testp=0
    if(g_a<=gmax_a)then
     gmin_b1=minval(kg(dim_b,ikg1:ikgn), MASK= (kg(dim_a,ikg1:ikgn)==g_a) )
     gmax_b1=maxval(kg(dim_b,ikg1:ikgn), MASK= (kg(dim_a,ikg1:ikgn)==g_a) )
     testp=1
    endif

!   Determine the boundaries for the plane -g_a or -g_a-1
    testm=0
    if( istwf_k==2 .or. istwf_k==3 .or. istwf_k==6 .or. istwf_k==7 )then
     if(-g_a>=gmin_a)then
      gmax_b2=minval(kg(dim_b,ikg1:ikgn), MASK= (kg(dim_a,ikg1:ikgn)==-g_a) )
      gmin_b2=maxval(kg(dim_b,ikg1:ikgn), MASK= (kg(dim_a,ikg1:ikgn)==-g_a) )
      testm=1
     endif
    else if( istwf_k==4 .or. istwf_k==5 .or. istwf_k==8 .or. istwf_k==9 )then
     if(-g_a-1>=gmin_a)then
      gmax_b2=minval(kg(dim_b,ikg1:ikgn), MASK= (kg(dim_a,ikg1:ikgn)==-g_a-1))
      gmin_b2=maxval(kg(dim_b,ikg1:ikgn), MASK= (kg(dim_a,ikg1:ikgn)==-g_a-1))
      testm=1
     endif
    endif
!   Must invert the boundaries, to use them for plane g_a
    if(testm==1)then
     if(istwf_k<=5)then
      gmax_b2=-gmax_b2 ; gmin_b2=-gmin_b2
     else
      gmax_b2=-gmax_b2-1 ; gmin_b2=-gmin_b2-1
     endif
    endif

    if( testp==1 .and. testm==1)then
     gmin_b=min(gmin_b1,gmin_b2) ; gmax_b=max(gmax_b1,gmax_b2)
    else if( testp==1 )then
     gmin_b=gmin_b1 ; gmax_b=gmax_b1
    else if( testm==1 )then
     gmin_b=gmin_b2 ; gmax_b=gmax_b2
    endif

!  Endif take into account time-reversal symmetry
   endif

   if (igb+1>2*mgfft+4) then
    write(message, '(a,a,a,a,3i8,a,a)' ) ch10,&
&     ' boundy: BUG -',ch10,&
&     '  iloop,igb,mgb=',iloop,igb,(2*mgfft+4),ch10,&
&     '  about to overwrite gbound.'
    call wrtout(06,message,'PERS')
    call leave_new('PERS')
   end if

   gbound(igb)=gmin_b
   gbound(igb+1)=gmax_b

   if( iloop==1 .and. istwf_k>=2 .and. istwf_k<=5 &
&                              .and. fftalgc==2 .and. g_a==0)then
!   If k_y=0 , for fftalgc==2, the g_a==0 plane must be completed
    if(istwf_k==2 .or. istwf_k==4)then
     gbound(igb+1)=max(gmax_b,-gmin_b)
     gbound(igb)=-gbound(igb+1)
    else if(istwf_k==3 .or. istwf_k==5)then
     gbound(igb+1)=max(gmax_b,-gmin_b-1)
     gbound(igb)=-gbound(igb+1)-1
    endif

   endif

   igb=igb+2

! Loop over g_a
  enddo

!Loop ii=1,2
 endif
 enddo 
!
 ngb=igb-1

!Debug mode
 if(fftalg>=0 .and. fftalg<=99)then
  write(6, '(a)' )' boundy : list of plane waves coordinates for 1st k point '
  write(6, '(a)' )'       ipw       kg(1:3,ipw) '
  do ipw=1,npw
   write(6, '(i10,a,3i6)' )ipw,'  ',kg(1:3,ipw+ikg)
  enddo
  write(6, '(a)' )' boundy : list of boundaries '
  write(6, '(a)' )'       igb     gbound(igb:igb+1) '
  do igb=1,ngb,2
   write(6, '(i10,a,2i6)' )igb,'  ',gbound(igb),gbound(igb+1)
  enddo
  stop
 endif

 end subroutine
!!***
