!{\src2tex{textfont=tt}}
!!****f* ABINIT/ccfft
!! NAME
!! ccfft
!!
!! FUNCTION
!! Carry out complex-to-complex fourier transforms between real
!! and reciprocal (G) space.
!! Library of such routines.
!! Include machine-dependent F90 routines used with fftalg=200.
!!
!! COPYRIGHT
!! Copyright (C) 2000-2001 ABINIT group (PT, XG, FF)
!! This file is distributed under the terms of the
!! GNU General Public License, see ~ABINIT/Infos/copyright
!! or http://www.gnu.org/copyleft/gpl.txt .
!!
!! INPUTS
!!  fftalga=govern the choice of the fft routine to be used
!!    if 1: SGoedecker routine
!!    if 2: Machine dependent routine, depending on the precompilation options
!!    if 3: FFTW library routine
!!  fftcache=size of the cache (kB)
!!  isign= Integer specifying which sign to be used for the transformation.
!!         must be either +1 or -1. 
!!  n1,n2,n3=Actual integer dimensions (see ngfft) for the 3D sequence.
!!           Physical dimension of the transform.
!!  n4,n5,n6=Leading dimensions. Generally, n6 is not different to n3.
!!  work1(2,n4,n5,n6)=Array to be transformed.
!!
!! OUTPUT
!!  inplace = 0 if result is in work2 ; =1 if result is in work1 (machine dependent)
!!  normalized=0 if the backward (isign=-1) FFT is not normalized, so has
!!                      to be normalized outside of ccfft 
!!            =1 otherwise
!!  work2(2,n4,n5,n6)=transformed array in case inplace=0.
!!
!! SIDE EFFECTS
!!  work1(2,n4,n5,n6)=at input, array to be transformed
!!                    at output, transformed array (in case inplace=1)
!!
!! NOTES
!! precompilation definitions :
!!   -D(machine_list) :  (case fftalga=200)
!!      choice of machine-dependent FFT library, if permitted 
!!   -DFFTW   : (case fftalga=300) activate the FFTW lib 
!!   -Dnolib  : (case fftalga=200) call SGoedecker routine, 
!!      instead of machine-dependent one
!!
!! More about fftalga=200
!! Library routines for the following platforms have been implemented :
!!  Compaq/DEC
!!  HP          (in place FFT)
!!  SGI         (in place FFT)
!!  NEC         (in place FFT)
!! For all the other platforms, or if the CPP directive nolib is
!! activated, one uses the fft routine from S. Goedecker.
!!
!! PARENTS
!!      fourdp,fourwf
!!
!! SOURCE

 subroutine ccfft(fftalga,fftcache,inplace,isign,normalized,&
& n1,n2,n3,n4,n5,n6,work1,work2)

 use basis_defs
 implicit none

!Arguments ------------------------------------
 integer, intent(in) :: fftalga,fftcache,isign,n1,n2,n3,n4,n5,n6   

 integer, intent(out) :: inplace,normalized              
 real(dp),dimension(2,n4,n5,n6),intent(out) :: work2 

 real(dp),dimension(2,n4,n5,n6),intent(inout):: work1  

!Local variables ------------------------------
 integer(i4b) :: done,ierr
 real(dp) :: ris,xnorm
 character*500 :: message

!Interfaces -----------------------------------
#if defined FFTW
 interface
  subroutine fftw(n1,n2,n3,isign,work1,work2)
   integer(i4b), intent(in) :: n1,n2,n3
   integer(i4b), intent(in) :: isign
   real(dp), dimension(:,:,:,:), intent(in) :: work1
   real(dp), dimension(:,:,:,:), intent(out) :: work2
  end subroutine fftw
 end interface
#endif

!Machine-dependent declarations---------------------------

#if defined dec_alpha && !defined nolib
!The following header should be included in any Fortran source file that
! calls ZFFT_3D: /usr/opt/XMDLOA331/dxml/DXMLDEF.FOR 
!Since this header is written in Fortran with old C-starting comments,
! compilation will fail with f90 -free
!This header comments should be first changed to !-starting by mean of the
! following command :
!     sed -e 's/^C/\!/' /usr/opt/XMDLOA331/dxml/DXMLDEF.FOR > DXMLDEF.free90
!
 include '../Machine_dept_files/DEC/DXMLDEF.free90'
!endif special header for dec_alpha
#endif

!---------------------------------------------------------

#if defined nec && !defined nolib
 integer :: nfftx
 integer :: ifax(60)
 real(dp), allocatable, save :: trigs(:)
#endif

!---------------------------------------------------------


#if defined sgi && !defined nolib
 integer,parameter :: nmax=15015
 integer,save :: n1_save,n2_save,n3_save
 integer(i4b) :: j1,j2,j3 ! local integers
 real(dp) :: xnorm ! Used to normalize the backward fft. 
 complex(dpc),dimension(nmax),save :: coeffs  ! Array of at least ( (N1+15)+(N2+15)+(N3+15) ) 
! elements.  On entry it contains the 
! Sines/Cosines and factorization of N. COEFFS 
! needs to be initialized with a call to zfft3di.
 complex(dpc),dimension(:,:,:),allocatable :: zarray ! Complex work array containing the samples 
! of the 3D sequence to be transformed.
 logical(lgt),save :: first=.true. ! logical used to initialize the coeffs calculation.
#endif

! *************************************************************************

 done=0

 if(fftalga==2)then

!---------------------------------------------------------

#if defined dec_alpha && !defined nolib 
  inplace=0 ; normalized=1
  if (isign==1) then
   ierr=zfft_3d('C','C','F',work1,work2,n1,n2,n3,n4,n5,1,1,1)
  else
   ierr=zfft_3d('C','C','B',work1,work2,n1,n2,n3,n4,n5,1,1,1)
  end if
  if (ierr /= 0) then
   write(message, '(a,a,a,a,i6,a,a)' ) ch10,&
&   ' ccfft : BUG -',ch10,&
&   '  error # ',ierr,' in calling zfft_3d',ch10
   call wrtout(06,message,'PERS')
   call leave_new('PERS')
  end if
  done=1
#endif

!---------------------------------------------------------

#if defined hp && !defined nolib
! The hp routine make the FFT in place
  inplace=1 ; normalized=1
  call Z3DFFT(work1,n1,n2,n3,n4,n5,isign,ierr)
! Check return code from Z3DFFT
  if(ierr /= 0)then
   write(message, '(a,a,a,a,i6,a,i6,a)' ) ch10,&
&   ' ccfft : BUG -',ch10,&
&   '  Z3DFFT isign=',isign,' error code =',ierr,'.'
   call wrtout(6,message,'PERS')
   call leave_new('PERS')
  endif
  done=1
#endif

!---------------------------------------------------------

#if defined nec && !defined nolib
! Here library ASL NEC routine
! Conventions for NEC FFT (see the ASL manual):
!  isign=1 implies a negative sign of the exponent
!  isign=-1 means a positive sign of the exponent
! Normalization is not included in the ASL routine
 
  allocate(trigs(2*(n4+n5+n6)))
  ifft = -isign
  nfftx = 2*n4*n5*n6
! The NEC routine makes the FFT in place ; normalisation not included
  inplace=1 ; normalized=0
  call ZFC3FB(n1,n2,n3,work1,n4,n5,n6,ifft,ifax,trigs,work2,ierr)
! Check return code from ZFC3FB
  if(ierr /= 0)then
   write(message, '(a,a,a,a,i6,a,i6,a)' )ch10,&
&   ' ccfft : BUG -',ch10,&
&   '  ZFC3FB isign=',ifft,' error code =',ierr,'.'
   call wrtout(6,message,'PERS')
   call leave_new('PERS')
  endif
  deallocate(trigs)
  done
#endif

!---------------------------------------------------------

#if defined sgi && !defined nolib

  if (first .or. n1/=n1_save .or. n2/=n2_save .or. n3/=n3_save) then
   if ((n1+n2+n3+45) > nmax) then
    write(message, '(a,a,a,a,a,a)' )ch10,&
&    ' ccfft : BUG -',ch10,&
&    '  n1+n2+n3+45>nmax.',ch10,&
&    '  Action : increase nmax in ccfft file.'
    call wrtout(6,message,'PERS')
    call leave_new('PERS')
   end if
   call zfft3di(n1,n2,n3,coeffs)
   first=.false. ; n1_save=n1 ; n2_save=n2 ; n3_save=n3
  end if

  inplace=1 ; normalized=0

!Original coding from P Thibaudeau (slightly modified)
! allocate(zarray(n4,n5,n6),stat=ierr)
! if (ierr /= 0) then
!  write(message, '(a,a,a,a,i6,a)' ) ch10,&
!&  ' ccfft : ERROR -',ch10,&
!&  '  error # :',ierr,' in executing allocate(zarray(n4,n5,n6)).'
!   call wrtout(06,message,'PERS')
!   call leave_new('PERS')
!  end if
!  do j3=1,n3
!   do j2=1,n2
!    do j1=1,n1
!     zarray(j1,j2,j3)=cmplx(work1(1,j1,j2,j3),work1(2,j1,j2,j3))
!    end do
!   end do
!  end do
!  call zfft3d(isign,n1,n2,n3,zarray,n4,n5,coeffs)
!  do j3=1,n3
!  do j2=1,n2
!   do j1=1,n1
!    work2(1,j1,j2,j3)=real(zarray(j1,j2,j3))
!    work2(2,j1,j2,j3)=imag(zarray(j1,j2,j3))
!   end do
!  end do
! end do
! deallocate(zarray,stat=ierr)
! end if

! XG 010414 Should work, and save CPU time
  call zfft3d(isign,n1,n2,n3,work1,n4,n5,coeffs)
  done=1

#endif

!---------------------------------------------------------

 else if(fftalga==3)then

#if defined FFTW
  inplace=0 ; normalized=0
  call fftw(n1,n2,n3,isign,work1,work2)  
  done=1
#endif

#if !defined FFTW
! If fftw is not accessible, one should not use fftalga=3
  write(message, '(a,a,a,a,a,a,a,a)' )ch10,&
&  ' ccfft : ERROR -',ch10,&
&  '  The library fftw has not been used to generate',ch10,&
&  '  the executable. Hence, fftalg(A)=3 is not allowed.',ch10,&
&  '  Action : check the value of fftalg in your input file.'
  call wrtout(6,message,'PERS')
  call leave_new('PERS')
#endif

!---------------------------------------------------------

 else if(fftalga<1 .or. fftalga>3)then

  write(message, '(a,a,a,a,a,a,i5,a,a)' )ch10,&
&  ' ccfft : ERROR -',ch10,&
&  '  The allowed values of fftalg(A) are 1, 2, 3 .',ch10,&
&  '  The actual value of fftalg(A) is',fftalga,ch10,&
&  '  Action : check the value of fftalg in your input file.'
  call wrtout(6,message,'PERS')
  call leave_new('PERS')

 endif

!---------------------------------------------------------

!In case fftalg is allowed, but the FFT has not yet been done 
 if(done==0)then
! Call Stefan Goedecker FFT
  inplace=0 ; normalized=0
  ris=real(isign,kind=dp)
  call sg_fft(fftcache,n4,n5,n6,n1,n2,n3,work1,work2,ris)
  ierr=0
 endif

 end subroutine ccfft

!!***
