!{\src2tex{textfont=tt}}
!!****f* ABINIT/fourdp
!!
!! NAME
!! fourdp
!! 
!! FUNCTION
!! Conduct fourier transform of REAL or COMPLEX function f(r)=fofr defined on
!! fft grid in real space, to create complex f(G)=fofg
!! defined on full fft grid in reciprocal space, in full storage mode,
!! or the reverse operation.
!! For the reverse operation, the final data is divided by nfft.
!! REAL case when cplex=1, COMPLEX case when cplex=2
!! Usually used for density and potentials.
!!
!! There are two different possibilities :
!!  fftalgb=0 means using the complex-to-complex FFT routine,
!!   irrespective of the value of cplex
!!  fftalgb=1 means using a real-to-complex FFT or a complex-to-complex FFT,
!!   depending on the value of cplex.
!!  The only real-to-complex FFT available is from SGoedecker library.
!! 
!! COPYRIGHT
!! Copyright (C) 1998-2001 ABINIT group (DCA, XG)
!! This file is distributed under the terms of the 
!! GNU General Public License, see ~ABINIT/Infos/copyright
!! or http://www.gnu.org/copyleft/gpl.txt .
!! For the initials of contributors, see ~ABINIT/Infos/contributors .
!!
!! INPUTS
!! isign=sign of Fourier transform exponent: current convention uses
!!  +1 for transforming from G to r, -1 for transforming from r to G.
!! nfft=number of fft grid points (=ngfft(1)*ngfft(2)*ngfft(3))
!! ngfft(1:3)=integer fft box dimensions
!! ngfft(4:6)=modified fft box dimensions to avoid cache thrashing
!! ngfft(7)=fftalg=choice of the algorithm
!! ngfft(8)=size of the cache (kB)
!!
!! SIDE EFFECTS
!! Input/Output
!! fofg(2,nfft)=f(G), complex.  
!! fofr(cplex*nfft)=input function f(r) (real or complex)
!!
!! TODO
!! work2 should be a pointer, and allocated inside ccfft,
!! if really needed ...
!!
!! PARENTS
!!      dieltcel,dieltcel7,dieltcel8,dieltcel9,gstate,hartre,initro,loper3
!!      mklocl,moddiel,prcref,setvtr,symrhg,tddft,vloca3,xcden,xcpot
!!
!! SOURCE

 subroutine fourdp(cplex,fofg,fofr,isign,nfft,ngfft)
 

 use basis_defs
 implicit none

!Arguments ------------------------------------
 integer :: cplex,isign,nfft
 integer :: ngfft(8)
 real(dp) :: fofg(2,nfft),fofr(cplex*nfft)

!Local variables-------------------------------
 integer, parameter :: mfac=11,mg=2048
 integer :: fftalg,fftalga,fftalgb,fftcache,ic1,ic2,ic3,ier,&
& ifft,index,inplace,i1,i2,i3,normalized,&
& n1,n1halfm,n1half1,n2,n2half1,&
& n3,n4,n4half1,n5,n5half1,n6
 integer :: aft1(mfac),bef1(mfac),ind1(mg),now1(mfac),&
& aft2(mfac),bef2(mfac),ind2(mg),now2(mfac),&
& aft3(mfac),bef3(mfac),ind3(mg),now3(mfac)
 real(dp) :: ris,xnorm
 real(dp) :: trig1(2,mg),trig2(2,mg),trig3(3,mg),tsec(2)
 real(dp), allocatable :: work1(:,:,:,:),work2(:,:,:,:),&
& wk2d_a(:,:,:,:),wk2d_b(:,:,:,:),wk2d_c(:,:,:,:),wk2d_d(:,:,:,:)

 character*500 :: message
 
! *************************************************************************
 
!Keep track of timing

 call timab(9,1,tsec)

 n1=ngfft(1) ; n2=ngfft(2) ; n3=ngfft(3)
 n4=ngfft(4) ; n5=ngfft(5) ; n6=ngfft(6)
 fftcache=ngfft(8) 
 fftalg=ngfft(7)
 fftalga=fftalg/100
 fftalgb=mod(fftalg,100)/10
 ris=dble(isign)
 xnorm=1.0d0/dble(nfft)

 if(fftalgb/=0 .and. fftalgb/=1) then
  write(message, '(a,a,a,a,i4,a,a,a,a,a)' )ch10,&
&  ' fourdp : ERROR -',ch10,&
&  '  The input algorithm number fftalg=',fftalg,' is not allowed.',ch10,&
&  '  The second digit (fftalg(B)) must be 0 or 1.',ch10,&
&  '  Action : change fftalg in your input file.'
  call wrtout(6,message,'PERS')
  call leave_new('PERS')
 endif

 if(fftalgb==1 .and. fftalga/=1)then
  write(message, '(a,a,a,a,i4,a,a,a,a,a)' )ch10,&
&  ' fourdp : ERROR -',ch10,&
&  '  The input algorithm number fftalg=',fftalg,' is not allowed.',ch10,&
&  '  When fftalg(B) is 1, the only allowed value for fftalg(A) is 1.',ch10,&
&  '  Action : change fftalg in your input file.'
  call wrtout(6,message,'PERS')
  call leave_new('PERS')
 endif

 if (n4<n1.or.n5<n2.or.n6<n3) then
  write(message, '(a,a,a,a,3i8,a,a,3i8,a)' ) ch10,&
&   ' fourdp: BUG -',ch10,&
&   '  Each of n4,n5,n6=',n4,n5,n6,ch10,&
&   '  must be >=      n1, n2, n3 =',n1,n2,n3,'.'
  call wrtout(06,message,'PERS')
  call leave_new('PERS')
 endif

!---------------------------------------------------------
!Here, one calls the complex-to-complex FFT subroutine
 if( fftalgb==0 .or. cplex==2 )then

  allocate(work1(2,n4,n5,n6),work2(2,n4,n5,n6))

  if(isign==1)then

!  Transfer fofg to the expanded fft box
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofg,n1,n2,n3,work1)
   do i3=1,n3
    do i2=1,n2
     index=n1*(i2-1+n2*(i3-1))
     do i1=1,n1
      work1(1,i1,i2,i3)=fofg(1,i1+index)
      work1(2,i1,i2,i3)=fofg(2,i1+index)
     enddo
    enddo
   enddo
!$OMP END PARALLEL DO

!  Actual 3D FFT
   call ccfft(fftalga,fftcache,inplace,isign,normalized,&
&   n1,n2,n3,n4,n5,n6,work1,work2)

!  Take data from expanded box and put it in the original box.
   if(cplex==1)then
!   REAL case

    if(inplace==0)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofr,n1,n2,n3,work2)
     do i3=1,n3
      do i2=1,n2
       index=n1*(i2-1+n2*(i3-1))
       do i1=1,n1
        fofr(i1+index)=work2(1,i1,i2,i3)
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    else if(inplace==1)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofr,n1,n2,n3,work1)
     do i3=1,n3
      do i2=1,n2
       index=n1*(i2-1+n2*(i3-1))
       do i1=1,n1
        fofr(i1+index)=work1(1,i1,i2,i3)
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    endif

   else
!   COMPLEX case

    if(inplace==0)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofr,n1,n2,n3,work2)
     do i3=1,n3
      do i2=1,n2
       index=2*n1*(i2-1+n2*(i3-1))
       do i1=1,n1
        fofr(2*i1-1+index)=work2(1,i1,i2,i3)
        fofr(2*i1  +index)=work2(2,i1,i2,i3)
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    else if(inplace==1)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofr,n1,n2,n3,work1)
     do i3=1,n3
      do i2=1,n2
       index=2*n1*(i2-1+n2*(i3-1))
       do i1=1,n1
        fofr(2*i1-1+index)=work1(1,i1,i2,i3)
        fofr(2*i1  +index)=work1(2,i1,i2,i3)
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    endif

   endif

  else if(isign==-1)then

!  Insert fofr into the augmented fft box
   if(cplex==1)then
!   REAL case
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofr,n1,n2,n3,work1)
    do i3=1,n3
     do i2=1,n2
      index=n1*(i2-1+n2*(i3-1))
      do i1=1,n1
!      copy data
       work1(1,i1,i2,i3)=fofr(i1+index)
       work1(2,i1,i2,i3)=0.0d0
      enddo
     enddo
    enddo
!$OMP END PARALLEL DO
   else
!   COMPLEX case
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofr,n1,n2,n3,work1)
    do i3=1,n3
     do i2=1,n2
      index=2*n1*(i2-1+n2*(i3-1))
      do i1=1,n1
!      copy data
       work1(1,i1,i2,i3)=fofr(2*i1-1+index)
       work1(2,i1,i2,i3)=fofr(2*i1  +index)
      enddo
     enddo
    enddo
!$OMP END PARALLEL DO
   endif ! cplex

!  Actual 3D FFT
   call ccfft(fftalga,fftcache,inplace,isign,normalized,&
&   n1,n2,n3,n4,n5,n6,work1,work2)

!  Transfer fft output to the original fft box
   if(normalized==0)then
    if(inplace==0)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofg,n1,n2,n3,work2)
     do i3=1,n3
      do i2=1,n2
       index=n1*(i2-1+n2*(i3-1))
       do i1=1,n1
        fofg(1,i1+index)=work2(1,i1,i2,i3)*xnorm
        fofg(2,i1+index)=work2(2,i1,i2,i3)*xnorm
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    else if(inplace==1)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofg,n1,n2,n3,work1)
     do i3=1,n3
      do i2=1,n2
       index=n1*(i2-1+n2*(i3-1))
       do i1=1,n1
        fofg(1,i1+index)=work1(1,i1,i2,i3)*xnorm
        fofg(2,i1+index)=work1(2,i1,i2,i3)*xnorm
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    endif
   else if(normalized==1)then
    if(inplace==0)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofg,n1,n2,n3,work2)
     do i3=1,n3
      do i2=1,n2
       index=n1*(i2-1+n2*(i3-1))
       do i1=1,n1
        fofg(1,i1+index)=work2(1,i1,i2,i3)
        fofg(2,i1+index)=work2(2,i1,i2,i3)
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    else if(inplace==1)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofg,n1,n2,n3,work1)
     do i3=1,n3
      do i2=1,n2
       index=n1*(i2-1+n2*(i3-1))
       do i1=1,n1
        fofg(1,i1+index)=work1(1,i1,i2,i3)
        fofg(2,i1+index)=work1(2,i1,i2,i3)
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    endif
   endif ! normalized==0 or 1

! Endif choice of isign
  endif

  deallocate(work1,work2)

!End simple algorithm
 endif

!---------------------------------------------------------
!Here sophisticated algorithm based on S. Goedecker routines,
!only for the REAL case.
!Take advantage of the fact that fofr is real, and that fofg
!has corresponding symmetry properties.

 if(fftalgb==1 .and. cplex==1)then

! Check that dimension is not exceeded
  if (n1>mg .or. n2>mg .or. n3>mg) then
   write(message, '(a,a,a,a,3i10,a,a,a,i10,a)' ) ch10,&
&   ' fourdp : BUG -',ch10,&
&   '  One of the dimensions n1,n2,n3=',n1,n2,n3,',',ch10,&
&   '  exceeds allowed dimension mg=',mg,'.'
   call wrtout(06,message,'PERS')
   call leave_new('PERS')
  endif

  n1half1=n1/2+1 ; n1halfm=(n1+1)/2
  n2half1=n2/2+1
! n4half1 or n5half1 are the odd integers >= n1half1 or n2half1
  n4half1=(n1half1/2)*2+1
  n5half1=(n2half1/2)*2+1

! The size of these arrays should be reduced.
!  allocate(wk2d_a(2,n4,n5,1),wk2d_b(2,n4,n5,1))
!  allocate(wk2d_c(2,2*n1halfm+1,n5,1),wk2d_d(2,2*n1halfm+1,n5,1))

! This sophisticated algorithm allows to decrease the memory needs.
  allocate(work1(2,n4,n5half1,n6),work2(2,n4,n5half1,n6))

  if(isign==1)then

!  Compute auxiliary arrays needed for FFTs, here forward FFT
   call sg_ctrig(n1,trig1,aft1,bef1,now1,1.0d0,ic1,ind1,mfac,mg)
   call sg_ctrig(n2,trig2,aft2,bef2,now2,1.0d0,ic2,ind2,mfac,mg)
   call sg_ctrig(n3,trig3,aft3,bef3,now3,1.0d0,ic3,ind3,mfac,mg)

!  Transfer fofg to the expanded fft box (only half of it)

!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) SHARED(fofg,n1,n2,n3,work1)
   do i3=1,n3
    do i2=1,n2half1
     index=n1*(i2-1+n2*(i3-1))
     do i1=1,n1
      work1(1,i1,i2,i3)=fofg(1,i1+index)
      work1(2,i1,i2,i3)=fofg(2,i1+index)
     enddo
    enddo
   enddo
!$OMP END PARALLEL DO

!$OMP PARALLEL DO SHARED(aft3,bef3,fftcache,ind3,ic3)&
!$OMP&SHARED(now3,n1,n2half1,n4,n5half1,n6,ris,trig3,work1,work2)&
!$OMP&PRIVATE(i2)
  do i2=1,n2half1
   call sg_fftz(fftcache,mfac,mg,n4,n5half1,n6,n1,i2,i2,work1,work2,&
&   trig3,aft3,now3,bef3,ris,ind3,ic3)
  enddo
!$OMP END PARALLEL DO

!  Loop over x-y planes

!$OMP PARALLEL PRIVATE(i1,i2,i3,index,wk2d_a,wk2d_b,wk2d_c,wk2d_d) &
!$OMP&SHARED(aft1,aft2,bef1,bef2,fftcache,fofg,fofr,ic1,ic2,ind1,ind2) &
!$OMP&SHARED(n1,n1half1,n1halfm,n2,n2half1,n3) &
!$OMP&SHARED(n4,n5,now1,now2,ris,trig1,trig2,work2)

  allocate(wk2d_a(2,n4,n5,1),wk2d_b(2,n4,n5,1))
  allocate(wk2d_c(2,2*n1halfm+1,n5,1),wk2d_d(2,2*n1halfm+1,n5,1))

!$OMP DO
   do i3=1,n3

    do i2=1,n2half1
     do i1=1,n1
      wk2d_c(1,i1,i2,1)=work2(1,i1,i2,i3)
      wk2d_c(2,i1,i2,1)=work2(2,i1,i2,i3)
     enddo
    enddo

    call sg_fftx(fftcache,mfac,mg,2*n1halfm+1,n5,1,n2half1,1,wk2d_c,wk2d_d,&
&    trig1,aft1,now1,bef1,ris,ind1,ic1)

!   Compute symmetric and antisymmetric combinations
    do i1=1,n1half1-1
     wk2d_a(1,i1,1,1)=wk2d_d(1,2*i1-1,1,1)
     wk2d_a(2,i1,1,1)=wk2d_d(1,2*i1  ,1,1)
    enddo
!   If n1 odd, must add last data
    if((2*n1half1-2)/=n1)then
     wk2d_a(1,n1half1,1,1)=wk2d_d(1,n1,1,1)
     wk2d_a(2,n1half1,1,1)=0.0d0
    endif
    do i2=2,n2half1
     do i1=1,n1half1-1
      wk2d_a(1,i1,i2,1)=      wk2d_d(1,2*i1-1,i2,1)-wk2d_d(2,2*i1,i2,1)
      wk2d_a(2,i1,i2,1)=      wk2d_d(2,2*i1-1,i2,1)+wk2d_d(1,2*i1,i2,1)
      wk2d_a(1,i1,n2+2-i2,1)= wk2d_d(1,2*i1-1,i2,1)+wk2d_d(2,2*i1,i2,1)
      wk2d_a(2,i1,n2+2-i2,1)=-wk2d_d(2,2*i1-1,i2,1)+wk2d_d(1,2*i1,i2,1)
     enddo
     if((2*n1half1-2)/=n1)then
      wk2d_a(1,n1half1,i2,1)=      wk2d_d(1,n1,i2,1)
      wk2d_a(2,n1half1,i2,1)=      wk2d_d(2,n1,i2,1)
      wk2d_a(1,n1half1,n2+2-i2,1)= wk2d_d(1,n1,i2,1)
      wk2d_a(2,n1half1,n2+2-i2,1)=-wk2d_d(2,n1,i2,1)
     endif
    enddo

    call sg_ffty(fftcache,mfac,mg,n4,n5,1,1,n1halfm,1,1,wk2d_a,wk2d_b,&
&    trig2,aft2,now2,bef2,ris,ind2,ic2)

!   Take real part data from expanded box and put it in the original box.
    do i2=1,n2
     index=n1*(i2-1+n2*(i3-1))
     do i1=1,n1half1-1
!    copy data
      fofr(2*i1-1+index)=wk2d_b(1,i1,i2,1)
      fofr(2*i1  +index)=wk2d_b(2,i1,i2,1)
     enddo
!    If n1 odd, must add last data
     if((2*n1half1-2)/=n1)then
      fofr(n1+index)=wk2d_b(1,n1half1,i2,1)
     endif
    enddo

!  End of loop over x-y planes
   enddo
!$OMP END DO
   deallocate(wk2d_a,wk2d_b,wk2d_c,wk2d_d)

!$OMP END PARALLEL

  else if(isign==-1)then

!  Compute auxiliary arrays needed for FFTs, here backward FFT
   call sg_ctrig(n1,trig1,aft1,bef1,now1,-1.0d0,ic1,ind1,mfac,mg)
   call sg_ctrig(n2,trig2,aft2,bef2,now2,-1.0d0,ic2,ind2,mfac,mg)
   call sg_ctrig(n3,trig3,aft3,bef3,now3,-1.0d0,ic3,ind3,mfac,mg)

!  Treat first x-transform in x-y plane, and multiply
!  by overall normalization factor 1/nfft

!  Loop over x-y planes

!$OMP PARALLEL PRIVATE(i1,i2,i3,index,wk2d_a,wk2d_b,wk2d_c,wk2d_d) &
!$OMP&SHARED(aft1,aft2,bef1,bef2,fftcache,fofr,ic1,ic2,ind1,ind2) &
!$OMP&SHARED(n1,n1half1,n1halfm,n2,n2half1,n3) &
!$OMP&SHARED(n4,n5,now1,now2,ris,trig1,trig2,work1,xnorm)

  allocate(wk2d_a(2,n4,n5,1),wk2d_b(2,n4,n5,1))
  allocate(wk2d_c(2,2*n1halfm+1,n5,1),wk2d_d(2,2*n1halfm+1,n5,1))
 
!$OMP DO
   do i3=1,n3
! write(6,*)'tache n :',OMP_GET_THREAD_NUM()
    do i2=1,n2
     index=n1*(i2-1+n2*(i3-1))
     do i1=1,n1half1-1
!     copy and normalize data
      wk2d_a(1,i1,i2,1)=fofr(2*i1-1+index)*xnorm
      wk2d_a(2,i1,i2,1)=fofr(2*i1  +index)*xnorm
     enddo
!    If n1 odd, must add last data
     if((2*n1half1-2)/=n1)then
      wk2d_a(1,n1half1,i2,1)=fofr(n1+index)*xnorm
      wk2d_a(2,n1half1,i2,1)=0.0d0
     endif
    enddo

    call sg_ffty(fftcache,mfac,mg,n4,n5,1,1,n1halfm,1,1,wk2d_a,wk2d_b,&
&    trig2,aft2,now2,bef2,ris,ind2,ic2)

!   Decompose symmetric and antisymmetric parts
    do i1=1,n1halfm
     wk2d_c(1,2*i1-1,1,1)=wk2d_b(1,i1,1,1)
     wk2d_c(2,2*i1-1,1,1)=0.0d0
     wk2d_c(1,2*i1,1,1)=wk2d_b(2,i1,1,1)
     wk2d_c(2,2*i1,1,1)=0.0d0
    enddo

    do i2=2,n2half1
     do i1=1,n1halfm
      wk2d_c(1,2*i1-1,i2,1)=(wk2d_b(1,i1,i2,1)+wk2d_b(1,i1,n2+2-i2,1))*0.5d0
      wk2d_c(2,2*i1-1,i2,1)=(wk2d_b(2,i1,i2,1)-wk2d_b(2,i1,n2+2-i2,1))*0.5d0
      wk2d_c(1,2*i1,i2,1)= (wk2d_b(2,i1,i2,1)+wk2d_b(2,i1,n2+2-i2,1))*0.5d0
      wk2d_c(2,2*i1,i2,1)=-(wk2d_b(1,i1,i2,1)-wk2d_b(1,i1,n2+2-i2,1))*0.5d0
     enddo
    enddo

    call sg_fftx(fftcache,mfac,mg,2*n1halfm+1,n5,1,n2half1,1,wk2d_c,wk2d_d,&
&    trig1,aft1,now1,bef1,ris,ind1,ic1)

    do i2=1,n2half1
     do i1=1,n1
       work1(1,i1,i2,i3)=wk2d_d(1,i1,i2,1)
       work1(2,i1,i2,i3)=wk2d_d(2,i1,i2,1)
     enddo
    enddo

   enddo
!$OMP END DO
  deallocate(wk2d_a,wk2d_b,wk2d_c,wk2d_d)
!$OMP END PARALLEL

!$OMP PARALLEL DO SHARED(aft3,bef3,fftcache,ind3,ic3)&
!$OMP&SHARED(now3,n1,n2half1,n4,n5half1,n6,ris,trig3,work1,work2)&
!$OMP&PRIVATE(i2)
  do i2=1,n2half1
   call sg_fftz(fftcache,mfac,mg,n4,n5half1,n6,n1,i2,i2,work1,work2,&
&   trig3,aft3,now3,bef3,ris,ind3,ic3)
  enddo
!$OMP END PARALLEL DO

!Transfer fft output to the original fft box

!$OMP PARALLEL DO PRIVATE(i1,i2,i3,index) &
!$OMP&SHARED(fofg,n1,n2,n2half1,n3,work2)
   do i3=1,n3
    do i2=1,n2half1
     index=n1*(i2-1+n2*(i3-1))
     do i1=1,n1
      fofg(1,i1+index)=work2(1,i1,i2,i3)
      fofg(2,i1+index)=work2(2,i1,i2,i3)
     enddo
    enddo
!   Complete missing values with complex conjugate
!   Inverse of ix is located at nx+2-ix , except for ix=1, for which it is 1.
    if(n2half1>2)then
     do i2=2,n2+1-n2half1
      index=n1*((n2+2-i2)-1)
      if(i3/=1)index=index+n1*n2*((n3+2-i3)-1)
      fofg(1,1+index)= work2(1,1,i2,i3)
      fofg(2,1+index)=-work2(2,1,i2,i3)
      do i1=2,n1
       fofg(1,n1+2-i1+index)= work2(1,i1,i2,i3)
       fofg(2,n1+2-i1+index)=-work2(2,i1,i2,i3)
      enddo
     enddo
    endif
   enddo
!$OMP END PARALLEL DO

! Endif choice of isign
  endif

!  deallocate(wk2d_a,wk2d_b,wk2d_c,wk2d_d,work1,work2)
  deallocate(work1,work2)

! DEBUG
! write(6,*)' fourdp : stop '  
! stop
! ENDDEBUG

 endif

 call timab(9,2,tsec)

 end subroutine
!!***
