!{\src2tex{textfont=tt}}
!!****f* ABINIT/fourwf
!!
!! NAME
!! fourwf
!! 
!! FUNCTION
!! Carry out composite fourier transforms between real and reciprocal (G) space.
!! Wavefunctions, contained in a sphere in reciprocal space,
!! can be FFT to real space. They can also be FFT from real space
!! to a sphere. Also, the density maybe accumulated, and a local
!! potential can be applied.
!! 
!! The different options are :
!! - reciprocal to real space and output the result (option=0),
!! - reciprocal to real space and accumulate the density (option=1) 
!! - reciprocal to real space, apply the local potential to the wavefunction
!!    in real space and produce the result in reciprocal space (option=2)
!! - real space to reciprocal space (option=3).
!!
!!    NOTE that in the latter case, fftalg=1x1 MUST be used. This
!!    may be changed in the future.
!!
!! The different sections of this routine corresponds to different
!! algorithms, used independently of each others :
!!(read first the description of the fftalg input variable in abinis_help)
!! - fftalg=xx0 : use simple complex-to-complex routines, without zero padding 
!!     (rather simple, so can be used to understand how fourwf.f works);
!! - fftalg=1x1 : use S Goedecker routines, with zero padding 
!!     (7/12 savings in execution time);
!! - fftalg=1x2 : call even more sophisticated coding also based on S Goedecker routines
!!
!! This routine contains many parts that differ only 
!! by small details, in order to treat each case with the better speed.
!! Also for better speed, it uses no F90 construct, except the allocate command
!! and for zeroing arrays.
!! 
!! COPYRIGHT
!! Copyright (C) 1998-2001 ABINIT group (DCA, XG, GMR, FF)
!! This file is distributed under the terms of the 
!! GNU General Public License, see ~ABINIT/Infos/copyright
!! or http://www.gnu.org/copyleft/gpl.txt .
!! For the initials of contributors, see ~ABINIT/Infos/contributors .
!!
!! INPUTS
!! cplex= if 1 , denpot is real, if 2 , denpot is complex
!!    (cplex=2 only allowed for option=2, and istwf_k=1)
!!    one can also use cplex=0 if option=0 or option=3
!! fofgin(2,npwin)=holds input wavefunction in G vector basis sphere.
!! gboundin(2*mgfft+4)=sphere boundary info for reciprocal to real space
!! gboundout(2*mgfft+4)=sphere boundary info for real to reciprocal space 
!! indpw_kin(4,npwin)=fft box index for reciprocal to real space
!! indpw_kout(4,npwout)=fft box index for real to reciprocal space
!! istwf_k=option parameter that describes the storage of wfs
!! npwin=number of elements in fofgin array (for option 0, 1 and 2)
!! npwout=number of elements in fofgout array (for option 2 and 3)
!! mgfft=maximum size of 1D FFTs
!! ngfft(1:3)=integer fft box dimensions
!! n4,n5,n6=ngfft(4),ngfft(5),ngfft(6), dimensions of fofr.
!! ngfft(7)=fftalg=choice of the algorithm
!! ngfft(8)=size of the cache (kB)
!! option= if 0: do direct FFT
!!         if 1: do direct FFT, then sum the density
!!         if 2: do direct FFT, multiply by the potential, then do reverse FFT
!!         if 3: do reverse FFT only
!! tim_fourwf=timing code of the calling routine (can be set to 0 if not attributed)
!! weight=weight to be used for the accumulation of the density in real space
!!         (needed only when option=1)
!!
!! SIDE EFFECTS
!! Input/Output
!! for option==0, fofgin(2,npwin)=holds input wavefunction in G sphere;
!!                fofr(2,n4,n5,n6) contains the output Fourier Transform of fofgin;
!!                no use of denpot, fofgout and npwout. 
!! for option==1, fofgin(2,npwin)=holds input wavefunction in G sphere;
!!                denpot(cplex*n4,n5,n6) contains the input density at input,
!!                and the updated density at output (accumulated);
!!                no use of fofgout and npwout.
!! for option==2, fofgin(2,npwin)=holds input wavefunction in G sphere;
!!                denpot(cplex*n4,n5,n6) contains the input local potential;
!!                fofgout(2,npwout) contains the output function; 
!! for option==3, fofr(2,n4,n5,n6) contains the input real space wavefunction;
!!                fofgout(2,npwout) contains its output Fourier transform;
!!                no use of fofgin and npwin.
!!
!! PARENTS
!!      cgwf3,getghc,mkrho,mkrho3,outstat,susk_dyn,susk_dyn_pgg,suskmm_dyn
!!      tddft,vtowfk3
!!
!! SOURCE

  subroutine fourwf(cplex,denpot,fofgin,fofgout,fofr,&
&  gboundin,gboundout,indpw_kin,indpw_kout,&
&  istwf_k,mgfft,ngfft,npwin,npwout,n4,n5,n6,option,tim_fourwf,weight)

 
 use basis_defs
 implicit none                              
 
!Arguments ------------------------------------
 integer :: cplex,istwf_k,mgfft,npwin,npwout,n4,n5,n6,option,tim_fourwf
 integer :: gboundin(2*mgfft+4),gboundout(2*mgfft+4),&
& indpw_kin(4,npwin),indpw_kout(4,npwout),ngfft(8)
 real(dp) :: weight
 real(dp) :: denpot(cplex*n4,n5,n6),fofgin(2,npwin),&
&  fofgout(2,npwout),fofr(2,n4,n5,n6)

!Local variables-------------------------------
 integer :: fftalg,fftalga,fftalgc,fftcache,ier,ifft,ig,inplace,&
& isign,i1,i1inv,i2,i2inv,i3,i3inv,nfft,normalized,n1,n2,n3
 integer :: shiftg(3),symm(3,3)
 real(dp) :: fim,fre,ris,xnorm
 real(dp) :: tsec(2)
 real(dp), allocatable :: work1(:,:,:,:)
 character*500 :: message

! *************************************************************************
 
! DEBUG
!write(6,*)' fourwf : enter '
!write(6,*)fofgin(1,1),fofgin(2,1)
!write(6,*)fofgin(1,101),fofgin(2,101)
!stop
! ENDDEBUG
!
! Accumulate timing
 call timab(240+tim_fourwf,1,tsec)

 n1=ngfft(1) ; n2=ngfft(2) ; n3=ngfft(3) ; nfft=n1*n2*n3
 fftcache=ngfft(8)
 fftalg=ngfft(7)  
 fftalga=fftalg/100 ; fftalgc=mod(fftalg,10)

 if(fftalgc<0 .and. fftalgc>2)then
  write(message, '(a,a,a,a,i4,a,a,a,a,a)' )ch10,&
&  ' fourwf : ERROR -',ch10,&
&  '  The input algorithm number fftalg=',fftalg,' is not allowed.',ch10,&
&  '  The third digit, fftalg(C), must be 0, 1, or 2',ch10,&
&  '  Action : change fftalg in your input file.'
  call wrtout(6,message,'PERS')
  call leave_new('PERS')
 endif

 if(fftalgc/=0 .and. fftalga/=1) then
  write(message, '(a,a,a,a,i4,a,a,a,a,a)' )ch10,&
&  ' fourwf : ERROR -',ch10,&
&  '  The input algorithm number fftalg=',fftalg,' is not allowed.',ch10,&
&  '  The first digit must be 1 when the last digit is not 0.',ch10,&
&  '  Action : change fftalg in your input file.'
  call wrtout(6,message,'PERS')
  call leave_new('PERS')
 endif

 if( option<0 .or. option>3 )then
  write(message, '(a,a,a,a,i4,a,a,a)' )ch10,&
&  ' fourwf : BUG -',ch10,&
&  '  The option number',option,' is not allowed.',ch10,&
&  '  Only option=0, 1, 2 or 3 are allowed presently.'
  call wrtout(6,message,'PERS')
  call leave_new('PERS')
 endif

 if( option==1 .and. cplex/=1 )then
  write(message, '(a,a,a,a,a,a,i4,a)' )ch10,&
&  ' fourwf : BUG -',ch10,&
&  '  With the option number 1, cplex must be 1,',ch10,&
&  '  but it is cplex=',cplex,'.'
  call wrtout(6,message,'PERS')
  call leave_new('PERS')
 endif

 if( option==2 .and. (cplex/=1 .and. cplex/=2) )then
  write(message, '(a,a,a,a,a,a,i4,a)' )ch10,&
&  ' fourwf : BUG -',ch10,&
&  '  With the option number 1, cplex must be 1 or 2,',ch10,&
&  '  but it is cplex=',cplex,'.'
  call wrtout(6,message,'PERS')
  call leave_new('PERS')
 endif

 shiftg(:)=0
 symm(:,:)=0
 symm(1,1)=1 ; symm(2,2)=1 ; symm(3,3)=1

!Now choose the algorithm

!------------------------------------------------------------------
!Here, use routines that make forwards FFT separately of backwards FFT,
!in particular, usual 3DFFT library routines, called in ccfft.
 if(fftalgc==0 .or.  fftalgc==1 .or. (fftalgc==2 .and. option==3))then

  allocate (work1(2,n4,n5,n6))

  if(option/=3)then
!  Insert fofgin into the fft box (array fofr) :
   call sphere(fofgin,npwin,fofr,n1,n2,n3,n4,n5,n6,&
&   indpw_kin,istwf_k,1,shiftg,symm)

   isign=1
!  Fourier transform fofr (reciprocal to real space)
!  The output might be in work1 or fofr, depending on inplace
   if(fftalgc==0)then
!   Call usual 3DFFT library routines or SG simplest 
!   complex-to-complex routine
    call ccfft(fftalga,fftcache,inplace,isign,normalized,&
&    n1,n2,n3,n4,n5,n6,fofr,work1)
   else
!   Call SG routine, with zero padding
    call sg_fftpad(fftcache,mgfft,n4,n5,n6,n1,n2,n3,fofr,work1,1.0d0,gboundin)
    inplace=0 ; normalized=0
   endif
  endif

  if(option==0 .and. inplace==0)then
!  Must output in fofr
!$OMP PARALLEL DO PRIVATE(i1,i2,i3) SHARED(fofr,n1,n2,n3,work1)
    do i3=1,n3
     do i2=1,n2
      do i1=1,n1
       fofr(1,i1,i2,i3)=work1(1,i1,i2,i3)
       fofr(2,i1,i2,i3)=work1(2,i1,i2,i3)
      enddo
     enddo
    enddo
!$OMP END PARALLEL DO
  endif

! Note that if option==0 everything is alright already, the
! output is available in fofr.

  if(option==1)then
!  Accumulate density
   if(inplace==0)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3) SHARED(denpot,n1,n2,n3,weight,work1)
    do i3=1,n3
     do i2=1,n2
      do i1=1,n1
       denpot(i1,i2,i3)=denpot(i1,i2,i3)+&
&             weight*(work1(1,i1,i2,i3)**2+work1(2,i1,i2,i3)**2)
      enddo
     enddo
    enddo
!$OMP END PARALLEL DO
   else if(inplace==1)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3) SHARED(denpot,fofr,n1,n2,n3,weight)
    do i3=1,n3
     do i2=1,n2
      do i1=1,n1
       denpot(i1,i2,i3)=denpot(i1,i2,i3)+&
&             weight*(fofr(1,i1,i2,i3)**2+fofr(2,i1,i2,i3)**2)
      enddo
     enddo
    enddo
!$OMP END PARALLEL DO
   endif
  endif
! Done with option 1 ...

  if(option==2)then

!  Apply local potential
   if(cplex==1)then

    if(inplace==0)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3) SHARED(denpot,fofr,n1,n2,n3,work1)
     do i3=1,n3
      do i2=1,n2
       do i1=1,n1
        fofr(1,i1,i2,i3)=denpot(i1,i2,i3)*work1(1,i1,i2,i3)
        fofr(2,i1,i2,i3)=denpot(i1,i2,i3)*work1(2,i1,i2,i3)
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    else if(inplace==1)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3) SHARED(denpot,fofr,n1,n2,n3)
     do i3=1,n3
      do i2=1,n2
       do i1=1,n1
        fofr(1,i1,i2,i3)=denpot(i1,i2,i3)*fofr(1,i1,i2,i3)
        fofr(2,i1,i2,i3)=denpot(i1,i2,i3)*fofr(2,i1,i2,i3)
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    endif

   else if(cplex==2)then

    if(inplace==0)then
!$OMP PARALLEL DO PRIVATE(fre,fim,i1,i2,i3) SHARED(denpot,fofr,n1,n2,n3,work1)
     do i3=1,n3
      do i2=1,n2
       do i1=1,n1
        fre=work1(1,i1,i2,i3)
        fim=work1(2,i1,i2,i3)
        fofr(1,i1,i2,i3)=denpot(2*i1-1,i2,i3)*fre &
&                       -denpot(2*i1  ,i2,i3)*fim
        fofr(2,i1,i2,i3)=denpot(2*i1-1,i2,i3)*fim &
&                       +denpot(2*i1  ,i2,i3)*fre
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    else if(inplace==1)then
!$OMP PARALLEL DO PRIVATE(fre,fim,i1,i2,i3) SHARED(denpot,fofr,n1,n2,n3)
     do i3=1,n3
      do i2=1,n2
       do i1=1,n1
        fre=fofr(1,i1,i2,i3)
        fim=fofr(2,i1,i2,i3)
        fofr(1,i1,i2,i3)=denpot(2*i1-1,i2,i3)*fre &
&                       -denpot(2*i1  ,i2,i3)*fim
        fofr(2,i1,i2,i3)=denpot(2*i1-1,i2,i3)*fim &
&                       +denpot(2*i1  ,i2,i3)*fre
       enddo
      enddo
     enddo
!$OMP END PARALLEL DO
    endif ! inplace

   endif ! cplex=2

  endif ! option=2

! The data for option==2 or option==3 is now in fofr.

  if(option==2 .or. option==3)then

   isign=-1

   if(fftalgc==0)then
!   Call usual 3DFFT library routines or SG simplest
!   complex-to-complex routine
    call ccfft(fftalga,fftcache,inplace,isign,normalized,&
&    n1,n2,n3,n4,n5,n6,fofr,work1)
   else
!   Call SG routine, with zero padding
    call sg_fftpad(fftcache,mgfft,n4,n5,n6,n1,n2,n3,fofr,work1,-1.0d0,gboundout)
    inplace=0 ; normalized=0
   endif

   xnorm=1.d0/dble(nfft)

   if(normalized==0)then
    if(inplace==0)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,ig) &
!$OMP&SHARED(fofgout,indpw_kout,npwout,work1,xnorm)
     do ig=1,npwout
      i1=indpw_kout(1,ig) ; i2=indpw_kout(2,ig) ; i3=indpw_kout(3,ig)
      fofgout(1,ig)=work1(1,i1,i2,i3)*xnorm
      fofgout(2,ig)=work1(2,i1,i2,i3)*xnorm
     enddo
!$OMP END PARALLEL DO
    else if(inplace==1)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,ig) &
!$OMP&SHARED(fofgout,fofr,indpw_kout,npwout,xnorm)
     do ig=1,npwout
      i1=indpw_kout(1,ig) ; i2=indpw_kout(2,ig) ; i3=indpw_kout(3,ig)
      fofgout(1,ig)=fofr(1,i1,i2,i3)*xnorm
      fofgout(2,ig)=fofr(2,i1,i2,i3)*xnorm
     enddo
!$OMP END PARALLEL DO
    endif ! inplace
   else if(normalized==1)then
    if(inplace==0)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,ig) &
!$OMP&SHARED(fofgout,indpw_kout,npwout,work1)
     do ig=1,npwout
      i1=indpw_kout(1,ig) ; i2=indpw_kout(2,ig) ; i3=indpw_kout(3,ig)
      fofgout(1,ig)=work1(1,i1,i2,i3)
      fofgout(2,ig)=work1(2,i1,i2,i3)
     enddo
!$OMP END PARALLEL DO
    else if(inplace==1)then
!$OMP PARALLEL DO PRIVATE(i1,i2,i3,ig) &
!$OMP&SHARED(fofgout,fofr,indpw_kout,npwout)
     do ig=1,npwout
      i1=indpw_kout(1,ig) ; i2=indpw_kout(2,ig) ; i3=indpw_kout(3,ig)
      fofgout(1,ig)=fofr(1,i1,i2,i3)
      fofgout(2,ig)=fofr(2,i1,i2,i3)
     enddo
!$OMP END PARALLEL DO
    endif ! inplace

   endif ! normalized

  endif ! if option==2 or 3

  deallocate(work1)

 endif

!------------------------------------------------------------------
!Here, call more sophisticated specialized 3-dimensional fft 
!(zero padding as well as maximize cache reuse) based on S Goedecker routines.
!Specially tuned for cache architectures.
 if( fftalgc==2 .and. option/=3)then

! Note that the arguments are the same as for fourwf
  call sg_fftrisc(cplex,denpot,fofgin,fofgout,fofr,gboundin,gboundout,&
&  indpw_kin,indpw_kout,istwf_k,&
&  mgfft,ngfft,npwin,npwout,n4,n5,n6,option,weight)

 endif
!------------------------------------------------------------------

! DEBUG
!write(6,*)' fourwf : exit '
!if(.true.)stop
! ENDDEBUG

!Accumulate timing
 call timab(240+tim_fourwf,2,tsec)

 end subroutine
!!***
