!{\src2tex{textfont=tt}}
!!****f* ABINIT/indfft
!!
!! NAME
!! indfft
!! 
!! FUNCTION
!! Take the data for boundary and index of FFT (indpw), manipulate them
!! for convenient use in fourwf, and output them in indpw_k
!! and gbound.
!! 
!! COPYRIGHT
!! Copyright (C) 1998-2001 ABINIT group (XG,AR)
!! This file is distributed under the terms of the 
!! GNU General Public License, see ~ABINIT/Infos/copyright
!! or http://www.gnu.org/copyleft/gpl.txt .
!! For the initials of contributors, see ~ABINIT/Infos/contributors .
!!
!! INPUTS
!! iind=shift of location of data for indpw
!! indpw((mpw+(2*mgfft+4))*mkmem)=indirect indexing array
!!   for pw index in fft box+G sphere boundary
!! mgfft=maximum size of 1D FFTs
!! mkmem =maximum number of k points which can fit in core memory
!! mpw=maximum number of planewaves as dimensioned in calling routine
!! ngfft(1:3)=integer fft box dimensions, see getng for ngfft(4:8)
!! npw_k=number of G vectors in basis at this k point
!!
!! OUTPUT
!! gbound(2*mgfft+4)=sphere boundary data
!! indpw_k(4,npw_k)=array which gives fft box index for given basis sphere
!!   in a representation that is directly usable by fourwf.f!!
!!
!! PARENTS
!!      dyfnl3,energy,mkrho,mkrho3,outstat,scfcv,suscep_dyn,suscep_stat,tddft
!!      vtorho,vtorho3,wfconv
!!
!! SOURCE

 subroutine indfft(gbound,iind,indpw,indpw_k,&
& mgfft,mpw,mkmem,ngfft,npw_k) 


 

 use basis_defs
 implicit none

!Arguments ------------------------------------
 integer :: iind,mgfft,mpw,mkmem,npw_k
 integer :: gbound(2*mgfft+4),indpw((mpw+(2*mgfft+4))*mkmem),&
&  indpw_k(4,npw_k),ngfft(8)

!Local variables-------------------------------
 integer :: fftalg,g1,g2,ipw,ifftig,igb,index,&
&  i1,i2,j1,j2,n1,n2,n3,n4,n5,n6
 integer, allocatable :: index2d(:,:)
 character*500 :: message
 
! *************************************************************************
!
!DEBUG
!write(6,*)' indfft : enter, debug '
!ENDDEBUG

 n1=ngfft(1) ; n2=ngfft(2) ; n3=ngfft(3)
 n4=ngfft(4) ; n5=ngfft(5) ; n6=ngfft(6)
 fftalg=ngfft(7)

!Compute three-dimensional FFT index
!$OMP PARALLEL DO PRIVATE(ifftig,ipw,j1,j2) &
!$OMP&SHARED(iind,indpw,indpw_k,npw_k,n1,n2)
 do ipw=1,npw_k
  ifftig=indpw(ipw+iind)
! Decomposes ifftig as ifftig= i1+ n1*(i2-1 + n2*(i3-1))
!     with i1=1...n1 , i2=1...n2 , i3=1...n3
  j1=(ifftig-1)/n1
  j2=j1/n2
  indpw_k(1,ipw)=ifftig-n1*j1
  indpw_k(2,ipw)=j1-n2*j2+1
  indpw_k(3,ipw)=j2+1
 enddo
!$OMP END PARALLEL DO
 gbound(:)=indpw(1+npw_k+iind:(2*mgfft+4)+npw_k+iind)
!A special treatment is needed for mod(fftalg,10)==2
 if(mod(fftalg,10)==2)then
! First, generate a 2d index for each column of data
  allocate(index2d(n1,n2))
  index2d(:,:)=0
  index=1
  igb=3
  do g2=0,gbound(2)
   do g1=0,gbound(igb+1)
    index2d(g1+1,g2+1)=index
    index=index+1
   enddo
   if(gbound(igb)<=-1)then
    do g1=gbound(igb)+n1,n1-1
     index2d(g1+1,g2+1)=index
     index=index+1
    enddo
   endif
   igb=igb+2
  enddo
  if(gbound(1)<=-1)then
   do g2=gbound(1)+n2,n2-1
    do g1=0,gbound(igb+1)
     index2d(g1+1,g2+1)=index
     index=index+1
    enddo
    if(gbound(igb)<=-1)then
     do g1=gbound(igb)+n1,n1-1
      index2d(g1+1,g2+1)=index
      index=index+1
     enddo
    endif
    igb=igb+2
   enddo
  endif
!DEBUG
!write(6,*)' index2d '
!do i2=1,n2
! do i1=1,n1
!  if(index2d(i1,i2)/=0)write(6,*)i1,i2,index2d(i1,i2)
! enddo
!enddo
!ENDDEBUG

! The 2d index has been generated
! Now, contract indpw_k(1,ipw) and indpw_k(2,ipw) into indpw_k(4,ipw)
! indpw_k(1,ipw) and indpw_k(2,ipw) are used to hold inverse of index2d,
! and for them, the second index does not fill 1:npw . It is only
! the number of z-transform FFTs.
!$OMP PARALLEL DO PRIVATE(ipw,i1,i2) &
!$OMP&SHARED(index2d,indpw_k,npw_k)
  do ipw=1,npw_k
   i1=indpw_k(1,ipw)
   i2=indpw_k(2,ipw)
   indpw_k(4,ipw)=index2d(i1,i2)
  enddo
!$OMP END PARALLEL DO
!DEBUG
! write(6,*)' indfft : indpw,indpw_k(1:4) for each pw '
! write(6,*)' n1,n2,n3 ',n1,n2,n3
! do ipw=1,npw_k
!  write(6, '(6i6)' )ipw,indpw(ipw),indpw_k(1:4,ipw)
! enddo
! stop
!ENDDEBUG
  indpw_k(1:2,1:npw_k)=0
  do i1=1,n1
   do i2=1,n2
    index=index2d(i1,i2)
    if(index/=0)then
     indpw_k(1,index)=i1
     indpw_k(2,index)=i2
    endif
   enddo
  enddo

!DEBUG
! write(6,*)' indfft : indpw,indpw_k(1:4) for each pw '
! write(6,*)' n1,n2,n3 ',n1,n2,n3
! do ipw=1,npw_k
!  write(6, '(6i6)' )ipw,indpw(ipw),indpw_k(1:4,ipw)
! enddo
! stop
!ENDDEBUG

  deallocate(index2d)

 endif

 end subroutine
!!***
