#if (ARPREC_HAVE_STD)
#include <cstdio>
#include <cstring>
#include <cfloat>
#include <cmath>
#else
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
#endif

#include "mp/mpreal.h"
#include "mp/mpint.h"

#include "mpslq3.h"

int mpslq3(const matrix<mp_real> &x, matrix<mp_real> &rel, 
           const mp_real &eps, double gamma) {
	int n = x.size();
  int history_length = 8;
  int nw  = mp::mpwds;
  int nw2 = 15;
  int print_interval = 50;
  int check_interval = 10;
  int save_interval  = 10;
  mp_real t;
  mp_real teps = eps * 1.0e20;
  mp_real max_bound = 0.0;

  matrix<mp_real> y(n);
  matrix<mp_real> h(n, n-1);
  matrix<mp_real> b(n, n);

  SET_PREC(nw2);
  mp_real wteps = power(mp_real(2.0), 72 - 48 * nw2);
  mp_real weps = wteps * pow(2.0, -72);
  mp_real wmx1 = power(mp_real(2.0), 48 * nw2 - 72);
  mp_real wmx2 = power(mp_real(2.0), 48 * nw2);

  matrix<mp_real> wy(n);
  matrix<mp_real> wh(n, n-1);
  matrix<mp_real> wa(n, n);
  matrix<mp_real> wb(n, n);
  matrix<mp_real> whist(n, history_length);
  SET_PREC(nw);

  matrix<double> dy(n);
  matrix<double> da(n, n);
  matrix<double> db(n, n);
  matrix<double> dh(n, n-1);
  matrix<double> dhist(n, history_length);

  matrix<double> dsy(n);
  matrix<double> dsa(n, n);
  matrix<double> dsb(n, n);
  matrix<double> dsh(n, n-1);

  bool use_only_one_pair = false;

  int result = RESULT_CONTINUE;
  init_pslq(x, y, h, b);

  if (debug_level >= 3) {
    y.print("Initial y:");
    b.print("Initial B:");
    h.print("Initial H:");
  }

  pslq_iter = 0;
  if (debug_level >= 2) {
    if (result == RESULT_CONTINUE)
      printf("Iteration %5d: Starting MP iterations.\n", pslq_iter);
  }
  while (result == RESULT_CONTINUE) {

    result = init_pslq2(wy, wh, wa, wb, y, h, wteps, nw2);
		whist.zero();
    if (debug_level >= 3) {
      wy.print("Initial wy:");
      wb.print("Initial wB:");
      wh.print("Initial wH:");
    }
    if (debug_level >= 2) {
      if (result == RESULT_CONTINUE)
        printf("Iteration %5d: Starting MPM iterations.\n", pslq_iter);
    }
    while (result == RESULT_CONTINUE) {

      result = init_pslq2(dy, dh, da, db, wy, wh, 1.0e-10, 5);
			dhist.zero();
      if (debug_level >= 3) {
        dy.print("Initial dy:");
        db.print("Initial dB:");
        dh.print("Initial dH:");
      }
      if (result == RESULT_CONTINUE) {

        int iter_saved = pslq_iter;
        lq_decomp(n, n-1, dh);
        save_pslq(da, db, dh, dy, dsa, dsb, dsh, dsy);

        if (debug_level >= 3)
          dh.print("after LQ factorization: dh");

        if (debug_level >= 2)
          printf("Iteration %5d: Starting DP iterations.\n", pslq_iter);
        while (result == RESULT_CONTINUE) {

          if (pslq_iter - iter_saved >= save_interval) {
            if (debug_level >= 3)
              printf("Iteration %5d: DP save.\n", pslq_iter);
            save_pslq(da, db, dh, dy, dsa, dsb, dsh, dsy);
            iter_saved = pslq_iter;
          }

          pslq_iter++;
          if (debug_level >= 2) {
            if (debug_level >= 3 || pslq_iter % print_interval == 0)
              printf("Iteration %5d\n", pslq_iter);
          }

          result = iterate_mpslq2(gamma, da, db, dh, dy, use_only_one_pair);
          use_only_one_pair = false;

          if (debug_level >= 3) {
            dy.print("Updated dy: ");
            db.print("Updated dB: ");
            da.print("Updated dA: ");
            dh.print("Updated dH: ");
          }

          if (result == RESULT_CONTINUE) {
            int check_result = 
              check_history(dy, dhist, 1.0e-14);
            if (check_result == RESULT_DUPLICATE)
              use_only_one_pair = true;
            insert_history(dy, dhist);
          }

        } /* end while */

        if (result == RESULT_VERY_LARGE_VALUE) {
          /* Double precision iteration resulted in loss of
             precision.  Restore to previously saved data.  */
          if (debug_level >= 2) {
            printf("Iteration %5d: DP iteration aborted.\n", pslq_iter);
            printf("                 Reverting to iteration %d\n", iter_saved);
          }
          pslq_iter = iter_saved;
          save_pslq(dsa, dsb, dsh, dsy, da, db, dh, dy);
        }

      } /* end if */

      if (result == RESULT_LARGE_VALUE      ||
          result == RESULT_VERY_LARGE_VALUE ||
          result == RESULT_RELATION_FOUND) {
        /* Update MPM arrays. */
        if (debug_level >= 2) {
          printf("Iteration %5d: MPM update\n", pslq_iter);
        }
        int update_result = update_mpslq2(da, db, wa, wb, wh, wy, weps, wteps);
        int check_result  = check_size_pslq(wa, wb, wmx1, wmx2);

        mp_real u, v;
        matrix_minmax(wy, u, v);
        if (debug_level >= 2) {
          printf("Iteration %5d: Min, max of wy =", pslq_iter);
          print(u);
          printf(",");
          print(v);
          printf("\n");
        }

        if (check_result == RESULT_VERY_LARGE_VALUE) {
          fprintf(stderr, "ERROR: MPM Array too large.\n");
          exit(-1);
        }
        if (check_result != RESULT_CONTINUE)
          update_result = check_result;

        if (debug_level >= 3) {
          wy.print("Updated wy: ");
          wa.print("Updated wA: ");
          wb.print("Updated wB: ");
          wh.print("Updated wH: ");
        }

        if (update_result == RESULT_CONTINUE &&
            result != RESULT_VERY_LARGE_VALUE)
          result = RESULT_CONTINUE;
        else
          result = update_result;
      } /* end if */

      SET_PREC(nw2);
      if (result == RESULT_RANGE_TOO_LARGE ||
          result == RESULT_VERY_LARGE_VALUE) {
        if (debug_level >= 2) {
          printf("Iteration %5d: Performing MPM iterations.\n", pslq_iter);
        }
        /* Perform MPM iterations until we can start 
           double-precision iterations. */
        lq_decomp(n, n-1, wh);
        int iter_checked = pslq_iter;
        do {
          if (pslq_iter - iter_checked >= check_interval) {
            int check_result = check_pslq(wy, 1.0e-10);
            if (check_result == RESULT_CONTINUE) {
              if (debug_level >= 2)
                printf("Iteration %5d: Return to DP iterations.\n", pslq_iter);
              break;
            }
            iter_checked = pslq_iter;
          }

          pslq_iter++;
          result = iterate_mpslq2(gamma, wa, wb, wh, wy, 
                                  wteps, use_only_one_pair, wmx1, wmx2);
          use_only_one_pair = false;

          if (debug_level >= 2)
            printf("Iteration %5d\n", pslq_iter);
          if (debug_level >= 3) {
            wy.print("Updated wy: ");
            wb.print("Updated wB: ");
            wh.print("Updated wH: ");
          }

          if (result == RESULT_CONTINUE) {
            int check_result = 
              check_history(wy, whist, teps);
            if (check_result == RESULT_DUPLICATE)
              use_only_one_pair = true;
            insert_history(wy, whist);
          }

        } while (result == RESULT_CONTINUE);
      }
      SET_PREC(nw);

    } /* end while */

    if (result == RESULT_RANGE_TOO_LARGE ||
        result == RESULT_VERY_LARGE_VALUE) {
      fprintf(stderr, "MPM range too large.\n");
      exit(-1);
    }

    if (result == RESULT_LARGE_VALUE ||
        result == RESULT_PRECISION_EXHAUSTED ||
        result == RESULT_RELATION_FOUND) {
      /* Update the MP arrays. */
      if (debug_level >= 2) {
        printf("Iteration %5d: MP update\n", pslq_iter);
      }
      int update_result = update_mpslq(wa, wb, b, h, y, eps, teps);
      //debug_level = 3;
      mp_real u, v;
      matrix_minmax(y, u, v);
      if (debug_level >= 2) {
        printf("Iteration %5d: Min, max of y =", pslq_iter);
        print(u);
        printf(",");
        print(v);
        printf("\n");
      }
      if (debug_level >= 3) {
        y.print("Updated y: ");
        b.print("Updated B: ");
        h.print("Updated H: ");
      }

      /* Compute norm bound. */
      SET_PREC(nw2);
      lq_decomp(n, n-1, wh);
      mp_real bnd;
      bound_pslq(wh, bnd);
      if (bnd > max_bound)
        max_bound = bnd;
      SET_PREC(nw);

      if (debug_level >= 2) {
        printf("Iteration %5d: norm bound =", pslq_iter);
        print(bnd);
        printf(", max bound =");
        print(max_bound);
        printf("\n");
      }

      result = update_result;
    } /* end if */

  } /* end while */

  int jm = -1;
  if (result == RESULT_RELATION_FOUND) {
    mp_real u, v;

    /* Output final norm bound. */
    matrix_minmax(y, u, v);
    bound_pslq(h, t);
    printf("Relation detected at iteration %d\n", pslq_iter);
    printf("  min, max of y =");
    print(u);
    print(v);
    printf(", bound = ");
    print(t);
    printf("\n");

    /* Relation found.  Select the relation with smallest norm. */
    u = DBL_MAX;
    for (int j = 0; j < n; j++) {
      if (abs(y(j)) < eps) {
        t = 0.0;

        for (int i = 0; i < n; i++) {
          t += sqr(b(j, i));
        }
        t = sqrt(t);

        if (t < u) {
          u = t;
          jm = j;
        }
      }
    }

    if (jm < 0) {
      fprintf(stderr, "ERROR: Invalid index.\n");
      exit(-1);
    }

    for (int i = 0; i < n; i++) {
      rel(i) = b(jm, i);
    }

  }

  return result;
}

