#if (ARPREC_HAVE_STD)
#include <cstdio>
#include <cstring>
#include <cfloat>
#include <cmath>
#else
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
#endif

#include "mp/mpreal.h"
#include "mp/mpint.h"

#include "pslq2.h"

int pslq2(const matrix<mp_real> &x, matrix<mp_real> &rel, 
          const mp_real &eps, double gamma) {
  int n = x.size();
  int print_interval = 250;
  int check_interval = 10;
  int save_interval  = 10;
  mp_real t;
  mp_real teps = eps * 1.0e20;
  double max_bound = 0.0;

  matrix<mp_real> y(n);
  matrix<mp_real> h(n, n-1);
  matrix<mp_real> b(n, n);

  matrix<double> dy(n);
  matrix<double> da(n, n);
  matrix<double> db(n, n);
  matrix<double> dh(n, n-1);

  matrix<double> dsy(n);
  matrix<double> dsa(n, n);
  matrix<double> dsb(n, n);
  matrix<double> dsh(n, n-1);

  int result;
  init_pslq(x, y, h, b);
  result = reduce_pslq(h, y, b, eps);

  if (debug_level >= 3) {
    y.print("Initial y:");
    b.print("Initial B:");
    h.print("Initial H:");
  }

  pslq_iter = 0;
  while (result == RESULT_CONTINUE) {

    result = init_pslq2(dy, dh, da, db, y, h, 1.0e-10, 6);
    if (debug_level >= 3) {
      dy.print("Initial dy:");
      da.print("Initial da:");
      db.print("Initial db:");
      dh.print("Initial dh:");
    }
    if (result == RESULT_CONTINUE) {

      int iter_saved = pslq_iter;
      lq_decomp(n, n-1, dh);
      save_pslq(da, db, dh, dy, dsa, dsb, dsh, dsy);

      if (debug_level >= 3) {
        dh.print("after LQ factorization: dh");
      }

      while (result == RESULT_CONTINUE) {

        if (pslq_iter - iter_saved >= save_interval) {
          if (debug_level >= 3)
            printf("Iteration %6d: DP save.\n", pslq_iter);
          save_pslq(da, db, dh, dy, dsa, dsb, dsh, dsy);
          iter_saved = pslq_iter;
        }

        pslq_iter++;
        if (debug_level >= 2) {
          if (debug_level >= 3 || pslq_iter % print_interval == 0)
            printf("Iteration %6d\n", pslq_iter);
        }

        result = iterate_pslq2(gamma, da, db, dh, dy);

        if (debug_level >= 3) {
          dy.print("Updated dy: ");
          db.print("Updated dB: ");
          da.print("Updated dA: ");
          dh.print("Updated dH: ");
        }

      }

      if (result == RESULT_VERY_LARGE_VALUE) {
        /* Double precision iteration resulted in loss of precision.
           Restore to previous data.                                    */
        if (debug_level >= 2) {
          printf("Iteration %6d: DP iteration aborted.\n", pslq_iter);
          printf("                 Reverting to iteration %d\n", iter_saved);
        }
        pslq_iter = iter_saved;
        save_pslq(dsa, dsb, dsh, dsy, da, db, dh, dy);
      }
    }

    if (result == RESULT_LARGE_VALUE       ||
        result == RESULT_VERY_LARGE_VALUE  ||
        result == RESULT_RELATION_FOUND) {

      /* Update the MP arrays. */
      int update_result = update_pslq(da, db, b, h, y, eps, teps);
      if (debug_level >= 2) {
        printf("Iteration %6d: MP update\n", pslq_iter);
      }
      if (debug_level >= 3) {
        y.print("Updated y: ");
        b.print("Updated B: ");
        h.print("Updated H: ");
      }

      /* Compute norm bound. */
      lq_decomp(n, n-1, dh);
      double bnd;
      bound_pslq(dh, bnd);
      if (bnd > max_bound)
        max_bound = bnd;

      if (bnd < 1.0e-300) {
        fprintf(stderr, "ERROR: bound too small.\n");
        exit(-1);
      }

      if (debug_level >= 2) {
				mp_real u, v;
				matrix_minmax(y, u, v);

        printf("Iteration %6d: norm bound =", pslq_iter);
        print(bnd);
        printf(", max bound =");
        print(max_bound);
        printf("\n");
        printf("Iteration %6d: min, max y =", pslq_iter);
        print(u);
				printf(",");
				print(v);
				printf("\n");
      }

      if (update_result == RESULT_CONTINUE &&
          result != RESULT_VERY_LARGE_VALUE)
        result = RESULT_CONTINUE;
      else
        result = update_result;
    }

    if (result == RESULT_RANGE_TOO_LARGE ||
        result == RESULT_VERY_LARGE_VALUE) {
      /* Perform MP iterations until we can start 
         double-precision iterations. */
      lq_decomp(n, n-1, h);

      int iter_checked = pslq_iter;
      do {
        if (pslq_iter - iter_checked >= check_interval) {
          int check_result = check_pslq(y, 1.0e-10, 6);
          if (check_result == RESULT_CONTINUE) {
            if (debug_level >= 2)
              printf("Iteration %6d: Return to DP iterations.\n", pslq_iter);
            break;
          }
          iter_checked = pslq_iter;
        }

        pslq_iter++;
        result = iterate_pslq(gamma, y, h, b, eps, teps);
      } while (result == RESULT_CONTINUE);
    }

  } /* outer while */

  int jm = -1;
  if (result == RESULT_RELATION_FOUND) {
    mp_real u, v;

    /* Output final norm bound. */
    matrix_minmax(y, u, v);
    bound_pslq(h, t);
    printf("Relation detected at iteration %d\n", pslq_iter);
    printf("  min, max of y =");
    print(u);
    print(v);
    printf(", bound = ");
    print(t);
    printf("\n");

    /* Relation found.  Select the relation with smallest norm. */
    u = DBL_MAX;
    for (int j = 0; j < n; j++) {
      if (abs(y(j)) < eps) {
        t = 0.0;

        for (int i = 0; i < n; i++) {
          t += sqr(b(i, j));
        }
        t = sqrt(t);

        if (t < u) {
          u = t;
          jm = j;
        }
      }
    }

    if (jm < 0) {
      fprintf(stderr, "ERROR: Invalid index.\n");
      exit(-1);
    }

    for (int i = 0; i < n; i++) {
      rel(i) = b(i, jm);
    }

  }

  return result;
}

