module quadglobal
use mpmodule
implicit none
integer ndebug, ndigits, nerror, nquadl
end module

program tquadt

!   David H. Bailey      2002-08-28
!   This is the ARPREC Fortran-90 version.

!   This software is provided for research purposes only.  
!   Commercial usage requires license agreement.

!   This work was supported by the Director, Office of Science, Division
!   of Mathematical, Information, and Computational Sciences of the
!   U.S. Department of Energy under contract number DE-AC03-76SF00098.

!   This program demonstrates the quadrature routine 'quadt', which employs the
!   tranformation t <-- tanh (pi/2 * sinh (x)).  quadt is suitable for any
!   function that is continuous, infinitely differentiable and integrable on a
!   finite open interval.  It can also be used for certain integrals on infinite
!   intervals, by making a suitable change of variable -- see below.  While
!   this routine is not quite as efficient as quadg for functions that are
!   regular on a closed interval, it can be used for functions with an
!   integrable singularity at one or both of the endpoints.  Further, this
!   scheme has the advantage that function evaluations at one level are all
!   utilized at the next level, thus saving considerable computation.

!   The function(s) to be integrated is(are) defined in external function
!   subprogram(s) -- see the sample function subprograms below.  The name(s) of
!   the function subprogram(s) must be included in appropriate type and external
!   statements in the main program.

!   Note that an integral of a function on an infinite interval can be
!   converted to an integral on a finite interval by means of a suitable
!   change of variable.  Examples (here the notation "inf" means infinity):

!   Int_a^inf f(t) dt  =  Int_0^(1/a) f(1/t)/t^2 dt         [for a > 0]
!   Int_0^inf f(t) dt  =  Int_0^1 [f(t) + f(1/t)/t^2] dt

!   Inputs set in parameter statement below:
!   kdebug Debug level setting.  Default = 2.
!   ndp    Digits of precision.  May not exceed mpipl in file mpmod90.f.
!            In some cases, ndp must be significantly greater than the desired
!            tolerance in the result-- see the examples below.
!   neps   Log10 of the desired tolerance in the result (negative integer).
!   nq1    Max number of phases in quadrature routine; adding 1 increases
!            (possibly doubles) the number of accurate digits in the result,
!            but also roughly doubles the run time and memory.  nq1 > 2.
!   nq2    Quadrature table size.  Ordinarily set to 8 * 2^nq1 + 100.

use mpmodule
use quadglobal
implicit none
integer i, kdebug, ndp, neps, nq1, nq2
double precision second, tm0, tm1
type (mp_real) err, fun1, fun2, fun3, fun4, fun5, fun6, quadt, t1, t2, t3
parameter (kdebug = 2, ndp = 420, neps = -400, nq1 = 8, &
  nq2 = 8 * 2 ** nq1 + 100)
type (mp_real) wk(nq2), xk(nq2), x1, x2
external quadt, fun1, fun2, fun3, fun4, fun5, fun6

!   Perform multiprecision initialization.

call mpinit
if (ndp > mpipl .or. nq1 < 2) stop
call mpsetprec (ndp)
mpeps = mpreal (10.d0) ** neps
mpoud = ndp
ndebug = kdebug
ndigits = ndp
nerror = 0
nquadl = nq1
write (6, 1) ndigits, neps, nquadl
1 format ('Quadt test'/'Digits =',i6,'  Epsilon =',i6,'   Quadlevel =',i6)

!   Initialize quadrature tables.

tm0 = second ()
call initqt (nq1, nq2, wk, xk)
tm1 = second ()
if (nerror > 0) stop
write (6, 2) tm1 - tm0
2 format ('Quadrature initialization completed: cpu time =',f12.6)

!   Perform quadratures.

write (6, 3)
3 format (/'Problem 1: Int_0^(pi/4) sec(t) dt = log(1+sqrt(2))')
x1 = 0.d0
x2 = 0.25d0 * mppic
tm0 = second ()
t1 = quadt (fun1, x1, x2, nq1, nq2, wk, xk)
tm1 = second ()
write (6, 4) tm1 - tm0
4 format ('Quadrature completed: cpu time =',f12.6)
t2 = log (1.d0 + sqrt (mpreal (2.d0)))
t3 = t2 - t1
write (6, 5)
5 format ('result, correct, error =')
call mpwrite (6, t1, t2, t3)

goto 111

!   Problem 2 crashes for the time being.

write (6, 6)
6 format (/'Problem 2: Int_0^1 t*log(1+t) dt = 1/4')
x1 = 0.d0
x2 = 1.d0
tm0 = second ()
t1 = quadt (fun2, x1, x2, nq1, nq2, wk, xk)
tm1 = second ()
write (6, 4) tm1 - tm0
t2 = 0.25d0
t3 = t2 - t1
write (6, 5)
call mpwrite (6, t1, t2, t3)

111 continue

write (6, 7)
7 format (/'Problem 3: Int_0^1 t^2*arctan(t) dt = pi/12 - 1/6 + log(2)/6')
x1 = 0.d0
x2 = 1.d0
tm0 = second ()
t1 = quadt (fun3, x1, x2, nq1, nq2, wk, xk)
tm1 = second ()
write (6, 4) tm1 - tm0
t2 = mppic / 12.d0 - mpreal (1.d0) / 6.d0 + log (mpreal (2.d0)) / 6.d0
t3 = t2 - t1
write (6, 5)
call mpwrite (6, t1, t2, t3)

goto 222

!   Problem 4 crashes for the time being.

write (6, 8)
8 format (/'Problem 4: Int_0^1 log^2(t) dt = 2')
x1 = 0.d0
x2 = 1.d0
tm0 = second ()
t1 = quadt (fun4, x1, x2, nq1, nq2, wk, xk)
tm1 = second ()
write (6, 4) tm1 - tm0
t2 = 2.d0
t3 = t2 - t1
write (6, 5)
call mpwrite (6, t1, t2, t3)

222 continue

write (6, 9)
9 format (/'Problem 5: Int_0^(pi/2) sqrt(tan(t)) dt = pi*sqrt(2)/2')
x1 = 0.d0
x2 = 0.5d0 * mppic
tm0 = second ()
t1 = quadt (fun5, x1, x2, nq1, nq2, wk, xk)
tm1 = second ()
write (6, 4) tm1 - tm0
t2 = 0.5d0 * mppic * sqrt (mpreal (2.d0))
t3 = t2 - t1
write (6, 5)
call mpwrite (6, t1, t2, t3)

write (6, 10)
10 format (/'Problem 6: Int_0^inf 1/(e^t*sqrt(t)) dt = sqrt(pi)')
x1 = 0.d0
x2 = 1.d0
tm0 = second ()
t1 = quadt (fun6, x1, x2, nq1, nq2, wk, xk)
tm1 = second ()
write (6, 4) tm1 - tm0
t2 = sqrt (mppic)
t3 = t2 - t1
write (6, 5)
call mpwrite (6, t1, t2, t3)

stop
end

function fun1 (t)

!    fun1(t) = sec(t)

use mpmodule
use quadglobal
implicit none
type (mp_real) fun1, t

fun1 = 1.d0 / cos (t)
return
end

function fun2 (t)

!   fun2(t) = t * log(1+t)

use mpmodule
use quadglobal
implicit none
type (mp_real) fun2, t

fun2 = t * log (1.d0 + t)
return
end

function fun3 (t)

!   fun3(t) = t^2 * arctan(t)

use mpmodule
use quadglobal
implicit none
type (mp_real) fun3, t

fun3 = t ** 2 * atan (t)
return
end

function fun4 (t)

!   fun4(t) = log^2(t)

use mpmodule
use quadglobal
implicit none
type (mp_real) fun4, t

fun4 = log (t) ** 2
return
end

function fun5 (t)

!   fun5(t) = sqrt(tan(t))

use mpmodule
use quadglobal
implicit none
type (mp_real) fun5, t

fun5 = sqrt (tan (t))
return
end

function fun6 (t)
use mpmodule
use quadglobal
implicit none
type (mp_real) fun6, rt, t, t1, t2

!   fun6(t) =  1/(e^t*sqrt(t)) + 1/(e^(1/t)*sqrt(1/t)*t^2)
!   Note that Int_0^1 fun6(t) dt  =  Int_0^Inf 1/(e^t*sqrt(t)) dt.

t1 = 1.d0 / (exp(t) * sqrt (t)) 
if (t .gt. 1.d-5) then
  rt = 1.d0 / t
  t2 = 1.d0 / (exp (rt) * sqrt (rt) * t ** 2)
else
  t2 = 0.d0
endif
fun6 = t1 + t2

return
end

subroutine initqt (nq1, nq2, wk, xk)

!   This subroutine initializes the quadrature arrays xk and wk using the
!   tanh-sinh function:  x(t) = tanh (pi/2 * sinh (t)).  The argument nq2, 
!   which is the space allocated for wk and xk in the calling program, must
!   be at least 8 * 2^nq1 + 100.
!   David H Bailey      2002-08-28

use mpmodule
use quadglobal
implicit none
integer i, ierror, j, k, k1, nq1, nq2
type (mp_real) eps, h, p2, t1, t2, t3, t4, t5, wk(nq2), xk(nq2)

if (ndebug >= 1) then
  write (6, 1)
1 format ('initqt: tanh-sinh quadrature initialization')
endif

p2 = 0.5d0 * mppic
i = 0
h = 2.d0
eps = 0.5d0 ** 24 * mpeps

do k = 0, nq1
  k1 = min (k + 1, 2)
  h = 0.5d0 * h

  do j = 1, 1000000
    i = i + 1
    if (i .ge. nq2) goto 110
    t1 = (k1 * j - 1) * h

!   xk(i) = tanh (p2 * sinh (t1))
!   wk(i) = p2 * cosh (t1) / cosh (p2 * sinh (t1)) ** 2

!   The next five lines are equivalent to the above two, but faster.

    call mpcsshf (t1, t2, t3)
    t3 = p2 * t3
    call mpcsshf (t3, t4, t5)
    xk(i) = t5 / t4
    wk(i) = p2 * t2 / t4 ** 2
    call mpgetpar ('mpier', ierror)
    if (ierror > 0) goto 120
    if (xk(i) .eq. 1.d0 .or. wk(i) .lt. eps) goto 100
  enddo

100 continue

  xk(i) = 0.d0
  wk(i) = 0.d0
  if (ndebug >= 2) then
    write (6, 2) k, nq1
2   format ('initqt: phase =',i3,' out of',i3,' completed.')
  endif
enddo

goto 130

110 continue

write (6, 3) nq2
3 format ('initqt: table space parameter is too small; value =',i8)
nerror = 91
goto 130

120 continue

nerror = ierror + 100
write (6, 4) nerror
4 format ('initqt: error in quadrature initialization; code =',i5)

130  continue

return
end

function quadt (fun, x1, x2, nq1, nq2, wk, xk)

!   This routine computes the integral of the function in fundx on the interval
!   [x1, x2], with up to nq1 iterations, with a target tolerance of mpeps.
!   wk and xk are precomputed tables of weights and abscissas, each of size
!   nq2 and of type mp_real.  The function fun is not evaluated at x1 or x2.

!   David H. Bailey     2002-08-28

use mpmodule
use quadglobal
implicit none
integer i, ierror, j, k, nds, nq1, nq2
double precision d1, d2, d3, d4, dplog10q
type (mp_real) a, b, c10, quadt, eps, eps1, eps2, err, fmx, fun, h, sum, &
  s1, s2, s3, t1, t2, t3, wk(nq2), xk(nq2), xx, x1, x2
external fun

nds = mpoud
mpoud = 56
a = 0.5d0 * (x2 - x1)
b = 0.5d0 * (x2 + x1)
xx = a * xk(1) + b
sum = wk(1) * fun (xx)
call mpgetpar ('mpier', ierror)
if (ierror > 0 .or. nerror > 0) goto 120
s1 = a * sum
s2 = 0.d0
i = 1
h = 2.d0
c10 = 10.d0
eps = mpeps
eps1 = 2.d0 ** 48 * mpeps

do k = 0, nq1
  h = 0.5d0 * h
  s3 = s2
  s2 = s1
  fmx = 1.d0

  do j = 1, 1000000000
    i = i + 1
    if (wk(i) .eq. 0.d0) goto 100
    xx = a * xk(i) + b
    if (xx < x2) then
      t1 = fun (xx)
      call mpgetpar ('mpier', ierror)
      if (ierror > 0 .or. nerror > 0) goto 120
    else
      t1 = 0.d0
    endif
    xx = - a * xk(i) + b
    if (xx > x1) then
      t2 = fun (xx)
      call mpgetpar ('mpier', ierror)
      if (ierror > 0 .or. nerror > 0) goto 120
    else
      t2 = 0.d0
    endif
    fmx = max (fmx, abs (t1), abs (t2))
    sum = sum + wk(i) * (t1 + t2)
  enddo

100 continue

!   Compute s1 = current integral approximation and err = error estimate.

  s1 =  a * h * sum
  eps2 = fmx * eps1
  d1 = dplog10q (abs (s1 - s2))
  d2 = dplog10q (abs (s1 - s3))
  d3 = dplog10q (eps2) - 10.d0

  if (k .le. 1) then
    err = 1.d0
  elseif (d1 .eq. -9999.d0) then
    err = 0.d0
  else
    d4 = min (0.d0, max (d1 ** 2 / d2, 2.d0 * d1, d3))
    err = c10 ** nint (d4)
  endif

!   Output current integral approximation and error estimate, to 56 dp.

  if (ndebug >= 2) then
    write (6, 1) k, nq1, nint (dplog10q (abs (err)))
1   format ('quadt: iteration',i3,' of',i3,'; est error = 10^',i5, &
      '; approx value =')
    call mpwrite (6, s1)
  endif
  if (k >= 3 .and. err < eps1) goto 130
  if (k >= 3 .and. err < eps2) goto 110
enddo

write (6, 3) nint (dplog10q (abs (err))), nquadl
3 format ('quadt: est error = 10^',i5/&
  'Increase Quadlevel for greater accuracy. Current Quadlevel =',i4)
goto 130

110 continue

write (6, 5) nint (dplog10q (abs (err))), ndigits
5 format ('quadt: est error = 10^',i5/&
  'Increase working prec (Digits) for greater accuracy. Current Digits =',i4)
goto 130

120 continue

if (ierror > 0) nerror = ierror + 100
write (6, 6) nerror
6 format ('quadt: error in quadrature calculation; code =',i5)
s1 = 0.d0

130 continue

quadt = s1
mpoud = nds
return
end

function dplog10q (a)

!   For input MP value a, this routine returns a DP approximation to log10 (a).

use mpmodule
implicit none
integer ia
double precision da, dplog10q, t1
type (mp_real) a

call mpmdc (a%mpr, da, ia)
if (da .eq. 0.d0) then
  dplog10q = -9999.d0
else
  dplog10q = log10 (abs (da)) + ia * log10 (2.d0)
endif

100 continue
return
end
