/*
 * src/mpcomplex.cc
 *
 * This software is provided for research purposes only.
 * Commercial usage requires license agreement.
 *
 * This work was supported by the Director, Office of Science, Division
 * of Mathematical, Information, and Computational Sciences of the
 * U.S. Department of Energy under contract number DE-AC03-76SF00098.
 *
 * Copyright (c) 2002
 *
 */
#include "mp/mpcomplex.h"

void mp_complex::mpcmuld(const mp_complex& a, double db, int n, 
			 mp_complex& b)
{
  mp_real::mpmuld(a.real, db, n, b.real);
  mp_real::mpmuld(a.imag, db, n, b.imag);  
}

void mp_complex::mpcadd(const mp_complex& a, const mp_complex& b, mp_complex& c)
{
  mp_real::mpadd(a.real, b.real, c.real);
  mp_real::mpadd(a.imag, b.imag, c.imag);
}

void mp_complex::mpcsub(const mp_complex& a, const mp_complex& b, mp_complex& c)
{
  mp_real::mpsub(a.real, b.real, c.real);
  mp_real::mpsub(a.imag, b.imag, c.imag);
  return;
}

void mp_complex::mpcmul(const mp_complex& a, const mp_complex& b, mp_complex& c)
{
  if(MPIER != 0) {
    if(MPIER == 99) mpabrt();
    zero(c);
    return;
  }
  if(MPIDB >= 7)
    cout << "\nMPCMUL I";

  // use 3 multiply method for complex multiplication.
  // This method requires more additions, but with this library,
  // multiplications are more expensive.
  mp_real temp1, temp2;
  mp_real::mpmul(a.real, b.real, temp1);
  mp_real::mpmul(a.imag, b.imag, temp2);
  mp_real::mpsub(temp1, temp2, c.real);
  mp_real::mpadd(temp2, temp1, temp2);
  mp_real temp3;
  mp_real::mpadd(a.real, a.imag, temp1);
  mp_real::mpadd(b.real, b.imag, temp3);
  mp_real::mpmul(temp1, temp3, temp3);
  mp_real::mpsub(temp3, temp2, c.imag);
  return;
}

void mp_complex::mpcmulx(const mp_complex& a, const mp_complex& b, mp_complex& c)
{
  if(MPIER != 0) {
    if(MPIER == 99) mpabrt();
    zero(c);
    return;
  }
  if(MPIDB >= 7)
    cout << "\nMPCMULX I";

  // use 3 multiply method for complex multiplication.
  // This method requires more additions, but with this library,
  // multiplications are more expensive.
  mp_real temp1, temp2;
  mp_real::mpmulx(a.real, b.real, temp1);
  mp_real::mpmulx(a.imag, b.imag, temp2);
  mp_real::mpsub(temp1, temp2, c.real);
  mp_real::mpadd(temp2, temp1, temp2);
  mp_real temp3;
  mp_real::mpadd(a.real, a.imag, temp1);
  mp_real::mpadd(b.real, b.imag, temp3);
  mp_real::mpmulx(temp1, temp3, temp3);
  mp_real::mpsub(temp3, temp2, c.imag);
  return;
}


void mp_complex::mpcdiv(const mp_complex& a, const mp_complex& b,
	    mp_complex& c)
{
  if(MPIER != 0) {
    if(MPIER == 99) mpabrt();
    zero(c);
    return;
  }
  if(MPIDB >= 8)
    cout << "\nMPCDIV I";

  if(b.real[1] == 0 && b.imag[1] == 0) {
    if(MPKER[16] != 0) {
      cout <<"\n*** MPCDIV: Divisor is zero";
      MPIER = 16;
      if(MPKER[MPIER] == 2) mpabrt();
    }
    return;
  }

  
  mp_real temp1, temp2, temp3, denom;
  mp_real::mpmul(a.real, b.real, temp1);
  mp_real::mpmul(a.imag, b.imag, temp2);
  //compute denominator and do not invert.
  mp_real::mpmul(b.real, b.real, temp3);
  mp_real::mpmul(b.imag, b.imag, denom);
  mp_real::mpadd(temp3, denom, denom);
  
  //real part of result
  mp_real::mpadd(temp1, temp2, temp3);
  mp_real::mpdiv(temp3, denom, c.real);
  
  //imag part of result
  mp_real::mpsub(temp2, temp1, temp2);
  mp_real::mpadd(a.real, a.imag, temp1);
  mp_real::mpsub(b.real, b.imag, temp3);
  mp_real::mpmul(temp1, temp3, temp3);
  mp_real::mpadd(temp2, temp3, temp3);
  mp_real::mpdiv(temp3, denom, c.imag);
}

void mp_complex::mpcdivx(const mp_complex& a, const mp_complex& b,
	    mp_complex& c)
{
  if(MPIER != 0) {
    if(MPIER == 99) mpabrt();
    zero(c);
    return;
  }
  if(MPIDB >= 8)
    cout << "\nMPCDIV I";

  if(b.real[1] == 0 && b.imag[1] == 0) {
    if(MPKER[18] != 0) {
      cout <<"\n*** MPCDIVX: Divisor is zero";
      MPIER = 18;
      if(MPKER[MPIER] == 2) mpabrt();
    }
    return;
  }

  //Check if precision level justifies advanced routine.
  int ncr = 1 <<mpmcrx;
  if(mpnw < ncr) {
    mpcdiv(a, b, c); return;
  }

  mp_real temp1, temp2, temp3, denom;
  mp_real::mpmulx(a.real, b.real, temp1);
  mp_real::mpmulx(a.imag, b.imag, temp2);
  //compute denominator and invert.
  mp_real::mpsqx(b.real, temp3);
  mp_real::mpsqx(b.imag, denom);
  mp_real::mpadd(temp3, denom, temp3);
  mp_real f(size_t(8));
  f[1] = 1.0; f[2] = 0.0; f[3] = 1.0; f[4] = 0.0;
  mp_real::mpdivx(f, temp3, denom);
  
  //real part of result
  mp_real::mpadd(temp1, temp2, temp3);
  mp_real::mpmulx(temp3, denom, c.real);
  
  //imag part of result
  mp_real::mpsub(temp1, temp2, temp2);
  mp_real::mpadd(a.real, a.imag, temp1);
  mp_real::mpsub(b.real, b.imag, temp3);
  mp_real::mpmulx(temp1, temp3, temp3);
  mp_real::mpsub(temp3, temp2, temp3);
  mp_real::mpmulx(temp3, denom, c.imag);
}

void mp_complex::mpceq(const mp_complex& a, mp_complex& b)
{
  mp_real::mpeq(a.real, b.real);
  mp_real::mpeq(a.imag, b.imag);
} 

void mp_complex::mpcsqx(const mp_complex& a, mp_complex& c)
{
  // use 3 multiply method for complex multiplication.
  // This method requires more additions, but with this library,
  // multiplications are more expensive.
  mp_real temp1, temp2;
  mp_real::mpsqx(a.real, temp1);
  mp_real::mpsqx(a.imag, temp2);
  mp_real::mpsub(temp1, temp2, c.real);
  mp_real::mpadd(temp2, temp1, temp2);
  mp_real::mpadd(a.real, a.imag, temp1);
  mp_real::mpsqx(temp1, temp1);
  mp_real::mpsub(temp1, temp2, c.imag);
  return;
}

void mp_complex::mpcagx(mp_complex& a, mp_complex& b)
{
  /**
   * This performs the arithmetic-geometric mean (AGM iterations.
   * This routine is called by MPANGX.  It is not intended to be called
   * directly by the uesr.
   */
  
  if(MPIER != 0) {
    if(MPIER == 99) mpabrt();
    zero(a);
    zero(b);
    return;
  }
  size_t n5 = (size_t)(mpnw+5);
  mp_complex sk0(n5), sk1(n5);
  int l1 = 0;
  sk0.real[1] = 1.0;
  sk0.real[2] = 0.0;
  double s1 = 2.0;
  
  //While condition checks for convergence
  while(l1<500 && sk0.real[1] != 0.0 && 
	(sk0.real[2] < s1 || sk0.real[2] >= -2)) {
    l1++;
    s1 = sk0.real[2];
    mpcadd(a, b, sk0);
    mpcmuld(sk0, 0.5, 0, sk1);
    mpcmulx(a, b, sk0);
    mpcsqrtx(sk0, b);
    mpceq(sk1, a);
    mp_real::mpsub(a.real, b.real, sk0.real);
  }
  if(l1 >= 500) {
    if(MPKER[12] != 0) {
      cout <<"\n*** MPCAGX: Iteration limit exceeded.";
      MPIER = 12;
      if(MPKER[MPIER] == 2) mpabrt();
    } 
    return;
  }
  if(MPIDB >= 6) cout << "\nMPCAGX: Iter, "<<l1<<
		   "\tTol. Achieved = " << sk0.real[2]; 
  return;
}

void mp_complex::mpcsqrtx(const mp_complex& a, mp_complex& b)
{
  /**
   * This routine computes the complex square root of the 
   * MPC number C. L is the offset between real and imaginary parts
   * in A and B.  L must be at least mpnw+4.  
   */

  if(MPIER != 0) {
    if(MPIER == 99) mpabrt();
    zero(b);
  }
  if(MPIDB >= 5) {
    cout <<"\nMPCSQRTX I";
  }

  if(a.real[1] == 0.0 && a.imag[1] == 0.0) {
    zero(b);
    return;
  }

  size_t n5 = mpnw + 5;
  mp_real sk0(n5), sk1(n5), sk2(n5);
  
  mp_real::mpsqx(a.real, sk0);
  mp_real::mpsqx(a.imag, sk1);
  mp_real::mpadd(sk0, sk1, sk2);
  mp_real::mpsqrtx(sk2, sk0);
  mp_real::mpeq(a.real, sk1);
  sk1[1] = ABS(sk1[1]); // take abs value.
  mp_real::mpadd(sk0, sk1, sk2);
  mp_real::mpmuld(sk2, 0.5, 0, sk1);
  mp_real::mpsqrtx(sk1, sk0);
  mp_real::mpmuld(sk0, 2.0, 0, sk1);
  if(a.real[1] >= 0.0) {
    mp_real::mpeq(sk0, b.real);
    mp_real::mpdivx(a.imag, sk1, b.imag);
  } else {
    mp_real::mpdivx(a.imag, sk1,  b.real);
    b.real[1] = ABS(b.real[1]);
    mp_real::mpeq(sk0, b.imag);
    b.imag[1] = SIGN(b.imag[1], a.imag[1]);
  }

  if(MPIDB >= 5) {
    cout <<"\nMPCSQRTX 0";
  }
  return;
}

void mp_complex::mpcpwx(const mp_complex& a, int n, mp_complex& b)
{
  /**
   * This computes the N-th power of the MPC number A and places
   * the MPC result in B.  When N is zero, 1 is returned.  When N is negative,
   * the reciprocal of A ^ |N| is returned.  Before calling MPNPWX, the
   * arrays of mpuu1 and uu2 must be initialized by calling mpinix.
   * For modest levels of precision, use MPNWPR.
   *
   * This routine emplys the normal binary method for exponentiation.
   */
  
  const double cl2 = 1.4426950408889633;

  if(MPIER != 0) {
    if(MPIER == 99) mpabrt();
    zero(b);
    return;
  }
  if(MPIDB >= 6) {
    cout << "\nMPCPWX I";
  }

  int ncr = 1 << mpmcrx;
  int na1 = MIN(int(ABS(a.real[1])), mpnw);
  int na2 = MIN(int(ABS(a.imag[1])), mpnw);
  
  // Check if precision level of A is too low to justify the advanced routine.
  
  if(na1 < ncr) {
    mpcpwr(a, n, b);
    return;
  }
  if(na1 == 0 && na2 == 0) {
    if(n >= 0) {
      zero(b);
      return;
    } else {
      // error : 0^N, N <= 0.
      if(MPKER[26] != 0) {
	cout <<"\n*** MPCPWX: argument is zero and N is negative or zero.";
	MPIER = 26;
	if(MPKER[MPIER] == 2) mpabrt();
      }
      return;
    }
  }

  size_t n5 = (size_t)(mpnw+5);
  int nn = ABS(n);
  mp_complex sk0(n5), sk1(n5), f1(size_t(8));
  //f1 = one.
  f1.real[1] = 1.0; f1.real[2] = 0.0; f1.real[3] = 1.0; f1.real[4] = 0.0;
  f1.imag[1] = 0.0; f1.imag[2] = 0.0; f1.imag[3] = 0.0; f1.imag[4] = 0.0;

  if(nn == 0) {
    mpceq(f1, b);
    return;
  } else if(nn == 1) {
    mpceq(a, b);
  } else if(nn == 2) {
    mpcsqx(a, b);
  } else {
    // full binary exponentiation.
    // Determine the least integer mn such that 2 & mn > nn.
    int mn = int(ANINT(cl2 * log(double(nn)) + 1.0 + mprxx));
    mpceq(a, sk0);
    mpceq(f1, b);
    int kn = nn, j, kk;

    for(j=1; j <= mn; j++) {
      kk = kn / 2;
      if(kn != 2 * kk) {
	mpcmulx(b, sk0, b);
      }
      kn = kk;
      if(j < mn)
	mpcsqx(sk0, sk0);
    }
  }
  
  // Compute the reciprocal if N is negative.
  if(n < 0) {
    mpcdivx(f1, b, sk0);
    mpceq(sk0, b);
  }
  return;
}

void mp_complex::mpcpwr(const mp_complex& a, int n, mp_complex& b)
{
  /**
   * This computes the N-th power of the MPC number A and places
   * the MPC result in B.  When N is zero, 1 is returned.  When N is negative,
   * the reciprocal of A ^ |N| is returned.  Before calling MPNPWX, the
   * arrays of mpuu1 and uu2 must be initialized by calling mpinix.
   * For modest levels of precision, use MPNWPR.
   *
   * This routine emplys the normal binary method for exponentiation.
   */
  
  const double cl2 = 1.4426950408889633;

  if(MPIER != 0) {
    if(MPIER == 99) mpabrt();
    zero(b);
    return;
  }
  if(MPIDB >= 7) {
    cout << "\nMPCPWR I";
  }

  int ncr = 1 << mpmcrx;
  int na1 = MIN(int(ABS(a.real[1])), mpnw);
  int na2 = MIN(int(ABS(a.imag[1])), mpnw);
  
  // Check if precision level of A is too low to justify the advanced routine.
  
  if(na1 < ncr) {
    mpcpwr(a, n, b);
    return;
  }
  if(na1 == 0 && na2 == 0) {
    if(n >= 0) {
      zero(b);
      return;
    } else {
      // error : 0^N, N <= 0.
      if(MPKER[25] != 0) {
	cout <<"\n*** MPCPWX: argument is zero and N is negative or zero.";
	MPIER = 25;
	if(MPKER[MPIER] == 2) mpabrt();
      }
      return;
    }
  }

  int nws = mpnw;
  mpnw++;
  size_t n5 = (size_t)(mpnw+5);
  int nn = ABS(n);
  mp_complex sk0(n5), sk1(n5), f1(size_t(8));
  //f1 = one.
  f1.real[1] = 1.0; f1.real[2] = 0.0; f1.real[3] = 1.0; f1.real[4] = 0.0;
  f1.imag[1] = 0.0; f1.imag[2] = 0.0; f1.imag[3] = 0.0; f1.imag[4] = 0.0;

  if(nn == 0) {
    mpceq(f1, b);
    mpnw = nws;
    return;
  } else if(nn == 1) {
    mpceq(a, b);
  } else if(nn == 2) {
    mpcsqx(a, b);
  } else {
    // full binary exponentiation.
    // Determine the least integer mn such that 2 & mn > nn.
    int mn = int(cl2 * log(double(nn)) + 1.0 + mprxx);
    mpceq(f1, b);
    mpceq(a, sk0);
    int kn = nn, j, kk;

    for(j=1; j <= mn; j++) {
      kk = kn / 2;
      if(kn != 2 * kk) {
	mpcmul(b, sk0, b);
      }
      kn = kk;
      if(j < mn)
	mpcsqx(sk0, sk0);
    }
  }
  
  // Compute the reciprocal if N is negative.
  if(n < 0) {
    mpcdiv(f1, b, sk0);
    mpceq(sk0, b);
  }
  mpnw = nws;
  mp_real::mproun(b.real);
  mp_real::mproun(b.imag);
  return;
}

void mp_complex::mpcsqrt(const mp_complex &a, mp_complex& b)
{
  /**
   * This routine computes the complex square root of the MPC number A.
   * for extra high levels of precision, use MPCSQRTX.
   * the last word of the result is not reliable. 
   *
   *
   * This routine uses the following formula, where A1 and A2 are
   * the real and imaginary parts of A, and where R = Sqrt [A1 ^2 + A2 ^ 2]:
   *
   * B = Sqrt [(R + A1) / 2] + I Sqrt [(R - A1) / 2]
   *
   * If the imaginary part of A is < 0, then the imaginary part of B
   * is also set to be < 0.
   */
  if(MPIER != 0) {
    if(MPIER == 99) mpabrt();
    zero(b);
  }
  if(MPIDB >= 6) {
    cout << "\nMPCSQRT I";
  }
  if(a.real[1] == 0 && a.imag[1] == 0) {
    zero(b);
    return;
  }
  size_t n6 = (size_t)(mpnw+6);
  mp_real sk0(n6), sk1(n6), sk2(n6);
  
  mp_real::mpmul(a.real, a.real, sk0);
  mp_real::mpmul(a.imag, a.imag, sk1);
  mp_real::mpadd(sk0, sk1, sk2);
  mp_real::mpsqrt(sk2, sk0);
  mp_real::mpeq(a.real, sk1);
  sk1[1] = ABS(sk1[1]);
  mp_real::mpadd(sk0, sk1, sk2);
  mp_real::mpmuld(sk2, 0.5, 0, sk1);
  mp_real::mpsqrt(sk1, sk0);
  mp_real::mpmuld(sk0, 2.0, 0, sk1);
  if(a.real[1] >= 0) {
    mp_real::mpeq(sk0, b.real);
    mp_real::mpdiv(a.imag, sk1, b.imag);
  } else {
    mp_real::mpdiv(a.imag, sk1, b.real);
    b.real[1] = ABS(b.real[1]);
    mp_real::mpeq(sk0, b.imag);
    b.imag[1] = SIGN(b.imag[1], a.imag[1]);
  }
  return;
}
