
#include "complextype.h"
#include "sincos.h"
#include "minmax.h"
#include "fxttypes.h"
#include "cmult.h"
#include "arith1.h"
//#include "newop.h"

#include "fxtio.h"

#include <cmath>


double
pythag(double a, double b)
// return sqrt(a*a+b*b) avoiding overflow.
// Shouldn't be necessary with type double and sane algorithms/code
{
    a = fabs(a);
    b = fabs(b);
    if ( a>b )
    {
        double v = b / a;
        return  a * sqrt(1.0+v*v);
    }
    else
    {
        if ( b==0.0 )  return 0.0;
        double v = a / b;
        return  b * sqrt(1.0+v*v);
    }
}
// -------------------------


double
extr3_estimate(const double *f)
// return extremum (y-value) of parabola through
// the points (-1,a), (0,b), (+1,c)
{
    double a = f[-1],  b = f[0],  c = f[1];
    double xval = 0.5*(a-c) / (a+c-b*2);   // position of maximum
    double yval = b - 0.25*(a-c)*xval;  // value of maximum
    return  yval;
}
// -------------------------

double
extr3_estimate(const double *f, double &xval)
// return extremum (y-value) of parabola through
// the points (-1,a), (0,b), (+1,c)
// Put get the corrsponding x-value into xval
{
    double a = f[-1],  b = f[0],  c = f[1];
    xval = 0.5*(a-c) / (a+c-b*2);   // position of maximum
    double yval = b - 0.25*(a-c)*xval;  // value of maximum
    return  yval;
}
// -------------------------


double
sinc(double x)
// return sin(pi*x)/(pi*x)
{
    if ( x==0 )  return  1.0;
    else
    {
        x *= M_PI;
        return  sin(x) / x;
    }
}
// -------------------------

void
chop(double &x, double eps=1e-6)
// set to zero if tiny
{
    if ( fabs(x)<eps )  x = 0.0;
}
// -------------------------

void
chop(double *f, ulong n, double eps/*=1e-6*/)
// set tiny elements to zero
{
    while ( n-- )  if ( fabs(f[n])<eps )  f[n] = 0.0;
}
// -------------------------


double
norm(const double *f, ulong n)
// return sqrt(\sum_k{f[k]*f[k]})
{
    double v = 0.0;
    while ( n-- )  v += (f[n]*f[n]);
    return  sqrt(v);
}
// -------------------------


double
normalize(double *f, ulong n, double v/*=1.0*/)
// scale f[] to norm v
{
    double  a = norm(f, n);
    if ( a!=0.0 )
    {
        v /= a;
        while ( n-- )  f[n] *= v;
    }
    return  a;
}
// -------------------------


double
sigma(const double *f, ulong n, double *mp/*=0*/)
// return standard deviation
// = sqrt(\sum_k{sqr(f[k]-mean(f[]))}/n)
{
    double m;
    if ( 0!=mp )  m = *mp;
    else          m = mean(f,n);

    double s = 0;
    for (ulong i=0; i<n; ++i)
    {
        double t = f[i] - m;
        s += t*t;
    }
    return  sqrt(s/n);
}
// -------------------------

void
mean_sigma(const double *f, ulong n, double *mp, double *sp)
// set *mp to mean, sp *to standard deviation of f[]
{
    *mp = mean(f,n);
    *sp = sigma(f,n,mp);
}
// -------------------------


void
smooth(double *f, ulong n, ulong m/*=1*/)
//  apply (1/4,1/2,1/4) - filter
{
    if ( n<=2 )  return;

    for (ulong k=0; k<m; ++k)
    {
        double t1 = f[0], t2;
        for (ulong i=1; i<=n-2; ++i)
        {
            t2 = t1;
            t1 = 0.25*(f[i-1]+f[i+1]) + 0.5*f[i];
            f[i-1] = t2;
        }
        f[n-2] = t1;
    }
}
// -------------------------


double
rms_diff(const double *f, const double *g, ulong n)
// return sqrt(\sum_k{sqr(f[k]-g[k])})/n
{
    double err = 0;
    for (ulong i=0; i<n; ++i)
    {
        double d = f[i]-g[i];
        err += d*d;
    }

    return  sqrt(err)/n;
}
// -------------------------


ulong *
histogram(const double *f, ulong n, ulong nbox, ulong *hh/*=0*/)
// put histogram of f[] (with nbox boxes) into hh[]
{
    ulong *h = hh;
    if ( 0==hh )  h = new ulong[nbox];

    for (ulong i=0; i<nbox; ++i)  h[i] = 0;

    double mi,ma;
    min_max(f, n, &mi, &ma);
    if ( ma==mi )
    {
        h[0] = n;
        return  h;
    }

    if ( ma != mi )
    {
        double v = 1.0*nbox/(ma-mi);
        for (ulong i=0; i<n; ++i)  ++h[(int)floor(v*(f[i]-mi))];
    }
    else  h[0] = n;

    return  h;
}
// -------------------------


void
welch_win(double *f, ulong n)
// multiply f[] with a welch window
{
    const double pin = M_PI/n;
    for (ulong i=0; i<n; i++)  f[i] *= (M_SQRT2*sin(i*pin));
}
// -------------------------


void
ri_multiply(const double *fr, const double *fi,
            double *gr, double *gi, ulong n)
// complex multiply array g[] by f[]:
//  Complex(gr[],gi[]) *= Complex(fr[],fi[])
{
    while ( n-- )  cmult(fr[n], fi[n], gr[n], gi[n]);
}
// -------------------------

