
#include "fxttypes.h"
#include "permutation.h"
#include "bitarray.h"


bool
is_identity(const ulong *f, ulong n)
//
// check whether f[] is the identical permutation,
// i.e. whether f[k]==k for all k= 0...n-1
//
{
    for (ulong k=0; k<n; ++k)  if ( f[k] != k )  return false;
    return true;
}
// -------------------------

bool
has_fixed_points(const ulong *f, ulong n)
//
// check whether f[] is a derangement of the trivial permutation
// whether f[k]==k for any k= 0...n-1
//
{
    for (ulong k=0; k<n; ++k)  if ( f[k] == k )  return true;
    return false;
}
// -------------------------


ulong
count_fixed_points(const ulong *f, ulong n)
//
// return number of fixed points in f[]
//
{
    ulong ct = 0;
    for (ulong k=0; k<n; ++k)  if ( f[k] == k )  ++ct;
    return ct;
}
// -------------------------


bool
is_derangement(const ulong *f, const ulong *g, ulong n)
//
// check whether f[] is a derangement of g[],
// i.e. whether f[k]!=g[k] for all k
//
{
    for (ulong k=0; k<n; ++k)  if ( f[k] == g[k] )  return false;
    return true;
}
// -------------------------

bool
is_derangement(const ulong *f, ulong n)
//
// check whether f[] is a derangement of identity
// i.e. whether f[k]!=k for all k
//
{
    for (ulong k=0; k<n; ++k)  if ( f[k] == k )  return false;
    return true;
}
// -------------------------


bool
is_valid_permutation(const ulong *f, ulong n, bitarray *bp/*=0*/)
//
// check whether all values 0...n-1 appear exactly once
//
{
    // check whether any element is out of range:
    for (ulong k=0; k<n; ++k)  if ( f[k]>=n )  return false;

    // check whether values are unique:
    bitarray *tp = bp;
    if ( 0==bp )  tp = new bitarray(n);  // tags
    tp->clear_all();

    ulong k;
    for (k=0; k<n; ++k)
    {
        if ( tp->test_set(f[k]) )  break;
    }

    if ( 0==bp )  delete tp;

    return  (k==n);
}
// -------------------------


bool
is_involution(const ulong *f, ulong n, bool hint/*=false*/)
//
// check whether max cycle length is <= 2
//
// set hint to true if the permutation is expected to be
// an involution
{
    if ( hint )
    {
        ulong z = 0;
        for (ulong k=0; k<n; ++k)  z |= (f[f[k]] ^ k);
        return  (0==z);
    }
    else
    {
        for (ulong k=0; k<n; ++k)  if ( f[f[k]] != k )  return false;
        return true;
    }
}
// -------------------------


bool
is_inverse(const ulong *f, const ulong *g, ulong n, bool hint/*=false*/)
//
// check whether f[] is inverse of g[]
//
// set hint to true if the permutations are expected to be
// mutual inverses
{
    if ( hint )
    {
        ulong z = 0;
        for (ulong k=0; k<n; ++k)  z |= (f[g[k]] ^ k);
        return  (0==z);
    }
    else
    {
        for (ulong k=0; k<n; ++k)  if ( f[g[k]] != k )  return false;
        return true;
    }
}
// -------------------------


bool
is_square(const ulong *f, const ulong *g, ulong n)
// whether f * f == g  as a permutation
{
    for (ulong k=0; k<n; ++k)  if ( g[k] != f[f[k]] )  return false;
    return true;
}
// -------------------------


ulong
permutation_index(const ulong *f, ulong n)
//
// Determine the permutation index for a given permutation list.
//
// The following function computes the ordinal of the given permutation,
// which is index of the permutation in sorting order:
//  1, 2, ..., n-1, n   is index 0
//  1, 2, ..., n, n-1   is index 1
//  ...
//  n, n-1, ..., 2, 1   is index n! -1
//
// The actual values of the elements are immaterial, only the relative
// ordering of the values is used.
//
// f[] is the array of elements of length n.
// The return value is the permutation index.
//
// note 1: complexity is n*n
// note 2: will overflow if index (which is <= n!-1) doesn't fit into a long
//  ==> only good for _tiny_ n
//.
// Original code by Thad Smith III, Boulder, CO  8/31/91
{
    ulong idx = 0;
    for (ulong i=1; i<n; i++)
    {
        ulong x = f[i-1];
        for (ulong j=i; j<n; j++)  if ( x>f[j] )  ++idx;
        idx *= n - i;
    }

    return idx;
}
// -------------------------
